from abc import ABC, abstractmethod
import logging
from queue import Queue
import threading
from typing import Collection, Optional, Type, Union

from platypush.common import StoppableThread
from platypush.context import get_bus
from platypush.entities.bluetooth import BluetoothDevice
from platypush.message.event.bluetooth import BluetoothDeviceEvent

from ._cache import EntityCache
from ._types import DevicesBlacklist, RawServiceClass


class BaseBluetoothManager(StoppableThread, ABC):
    """
    Abstract interface for Bluetooth managers.
    """

    def __init__(
        self,
        interface: str,
        poll_interval: float,
        connect_timeout: float,
        scan_lock: threading.RLock,
        scan_enabled: threading.Event,
        device_queue: Queue[BluetoothDevice],
        service_uuids: Optional[Collection[RawServiceClass]] = None,
        device_cache: Optional[EntityCache] = None,
        exclude_known_noisy_beacons: bool = True,
        blacklist: Optional[DevicesBlacklist] = None,
        **kwargs,
    ):
        """
        :param interface: The Bluetooth interface to use.
        :param poll_interval: Scan interval in seconds.
        :param connect_timeout: Connection timeout in seconds.
        :param scan_lock: Lock to synchronize scanning access to the Bluetooth device.
        :param scan_enabled: Event used to enable/disable scanning.
        :param device_queue: Queue used by the ``EventHandler`` to publish
            updates with the new parsed device entities.
        :param device_cache: Cache used to keep track of discovered devices.
        :param exclude_known_noisy_beacons: Exclude known noisy beacons.
        :param blacklist: Blacklist of devices to exclude from discovery.
        """
        from ._plugins import scan_plugins

        kwargs['name'] = f'Bluetooth:{self.__class__.__name__}'
        super().__init__(**kwargs)

        self.logger = logging.getLogger(__name__)
        self.poll_interval = poll_interval
        self._interface: Optional[str] = interface
        self._connect_timeout: float = connect_timeout
        self._service_uuids: Collection[RawServiceClass] = service_uuids or []
        self._scan_lock = scan_lock
        self._scan_enabled = scan_enabled
        self._device_queue = device_queue
        self._exclude_known_noisy_beacons = exclude_known_noisy_beacons
        self._blacklist = blacklist or DevicesBlacklist()

        self._cache = device_cache or EntityCache()
        """ Cache of discovered devices. """
        self._plugins = scan_plugins(self)
        """ Plugins compatible with this manager. """

    def notify(
        self, event_type: Type[BluetoothDeviceEvent], device: BluetoothDevice, **kwargs
    ):
        """
        Notify about a device update event by posting a
        :class:`platypush.message.event.bluetooth.BluetoothDeviceEvent` event on
        the bus and pushing the updated entity upstream.
        """
        get_bus().post(event_type.from_device(device=device, **kwargs))
        self._device_queue.put_nowait(device)

    @property
    def plugins(self):
        return self._plugins

    @abstractmethod
    def connect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
        interface: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        """
        Pair and connect to a device by address or name.

        :param device: The device address or name.
        :param port: The Bluetooth port to use.
        :param service_uuid: The service UUID to connect to.
        :param interface: The Bluetooth interface to use (it overrides the
            default ``interface``).
        :param timeout: The connection timeout in seconds (it overrides the
            default ``connect_timeout``).
        """

    @abstractmethod
    def disconnect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
    ):
        """
        Close an active connection to a device.

        :param device: The device address or name.
        :param port: If connected to a non-BLE device, the optional port to
            disconnect. Either ``port`` or ``service_uuid`` is required for
            non-BLE devices.
        :param service_uuid: The UUID of the service to disconnect from.
        """

    @abstractmethod
    def scan(self, duration: Optional[float] = None) -> Collection[BluetoothDevice]:
        """
        Scan for Bluetooth devices nearby and return the results as a list of
        entities.

        :param duration: Scan duration in seconds (default: same as the plugin's
            `poll_interval` configuration parameter)
        """

    @abstractmethod
    def read(
        self,
        device: str,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ) -> bytearray:
        """
        :param device: Name or address of the device to read from.
        :param service_uuid: Service UUID.
        :param interface: Bluetooth adapter name to use (default configured if None).
        :param connect_timeout: Connection timeout in seconds (default: same as the
            configured `connect_timeout`).
        """

    @abstractmethod
    def write(
        self,
        device: str,
        data: Union[bytes, bytearray],
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ):
        """
        :param device: Name or address of the device to read from.
        :param data: Raw data to be sent.
        :param service_uuid: Service UUID.
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param connect_timeout: Connection timeout in seconds (default: same as the
            configured `connect_timeout`).
        """


# vim:sw=4:ts=4:et:



import logging
from queue import Queue
from threading import Event
from typing import Callable, Optional, Sequence

from platypush.message.event.assistant import AssistantEvent, ConversationTimeoutEvent
from platypush.utils import wait_for_either

from ._intent import IntentProcessor
from ._stt import SttProcessor


class SpeechProcessor:
    """
    Speech processor class that wraps the STT and Intent processors under the
    same interface.
    """

    def __init__(
        self,
        stop_event: Event,
        model_path: Optional[str] = None,
        stt_enabled: bool = True,
        intent_enabled: bool = False,
        conversation_timeout: Optional[float] = None,
        get_cheetah_args: Callable[[], dict] = lambda: {},
        get_rhino_args: Callable[[], dict] = lambda: {},
    ):
        self.logger = logging.getLogger(self.__class__.__name__)
        self._stt_enabled = stt_enabled
        self._intent_enabled = intent_enabled
        self._model_path = model_path
        self._conversation_timeout = conversation_timeout
        self._audio_queue = Queue()
        self._stop_event = stop_event
        self._get_cheetah_args = get_cheetah_args
        self._get_rhino_args = get_rhino_args

        self._stt_processor = SttProcessor(
            conversation_timeout=conversation_timeout,
            stop_event=stop_event,
            enabled=stt_enabled,
            get_cheetah_args=get_cheetah_args,
        )

        self._intent_processor = IntentProcessor(
            conversation_timeout=conversation_timeout,
            stop_event=stop_event,
            enabled=intent_enabled,
            get_rhino_args=get_rhino_args,
        )

    @property
    def enabled(self) -> bool:
        """
        The processor is enabled if either the STT or the Intent processor are
        enabled.
        """
        return self._stt_enabled or self._intent_enabled

    def should_stop(self) -> bool:
        return self._stop_event.is_set()

    def next_event(self, timeout: Optional[float] = None) -> Optional[AssistantEvent]:
        evt = None

        # Wait for either the STT or Intent processor to finish processing the audio
        completed = wait_for_either(
            self._stt_processor.processing_done,
            self._intent_processor.processing_done,
            self._stop_event,
            timeout=timeout,
        )

        if not completed:
            self.logger.warning('Timeout while waiting for the processors to finish')

        # Immediately return if the stop event is set
        if self.should_stop():
            return evt

        with self._stt_processor._state_lock, self._intent_processor._state_lock:
            # Priority to the intent processor event, if the processor is enabled
            if self._intent_enabled:
                evt = self._intent_processor.last_event()

            # If the intent processor didn't return any event, then return the STT
            # processor event
            if (
                not evt or isinstance(evt, ConversationTimeoutEvent)
            ) and self._stt_enabled:
                # self._stt_processor.processing_done.wait(timeout=timeout)
                evt = self._stt_processor.last_event()

            if evt:
                self._stt_processor.reset()
                self._intent_processor.reset()

            return evt

    def process(
        self, audio: Sequence[int], block: bool = True, timeout: Optional[float] = None
    ) -> Optional[AssistantEvent]:
        """
        Process an audio frame.

        The audio frame is enqueued to both the STT and Intent processors, if
        enabled. The function waits for either processor to finish processing
        the audio, and returns the event from the first processor that returns
        a result.

        Priority is given to the Intent processor if enabled, otherwise the STT
        processor is used.
        """
        # Enqueue the audio to both the STT and Intent processors if enabled
        if self._stt_enabled:
            self._stt_processor.enqueue(audio)

        if self._intent_enabled:
            self._intent_processor.enqueue(audio)

        if not block:
            return None

        return self.next_event(timeout=timeout)

    def __enter__(self):
        """
        Context manager entry point - it wraps :meth:`start`.
        """
        self.start()

    def __exit__(self, *_, **__):
        """
        Context manager exit point - it wraps :meth:`stop`.
        """
        self.stop()

    def start(self):
        """
        Start the STT and Intent processors.
        """
        if self._stt_enabled:
            self._stt_processor.start()

        if self._intent_enabled:
            self._intent_processor.start()

    def stop(self):
        """
        Stop the STT and Intent processors.
        """
        self._stt_processor.stop()
        self._intent_processor.stop()

    def on_conversation_start(self):
        if self._stt_enabled:
            self._stt_processor.on_conversation_start()

        if self._intent_enabled:
            self._intent_processor.on_conversation_start()

    def on_conversation_end(self):
        if self._stt_enabled:
            self._stt_processor.on_conversation_end()

        if self._intent_enabled:
            self._intent_processor.on_conversation_end()

    def on_conversation_reset(self):
        if self._stt_enabled:
            self._stt_processor.on_conversation_reset()

        if self._intent_enabled:
            self._intent_processor.on_conversation_reset()

    def reset(self):
        """
        Reset the state of the STT and Intent processors.
        """
        self._stt_processor.reset()
        self._intent_processor.reset()

    @property
    def sample_rate(self) -> int:
        """
        The sample rate of the audio frames.
        """
        if self._intent_enabled:
            return self._intent_processor.sample_rate

        if self._stt_enabled:
            return self._stt_processor.sample_rate

        raise ValueError('No processor enabled')

    @property
    def frame_length(self) -> int:
        """
        The frame length of the audio frames.
        """
        if self._intent_enabled:
            return self._intent_processor.frame_length

        if self._stt_enabled:
            return self._stt_processor.frame_length

        raise ValueError('No processor enabled')



import inspect
import json


class ProcedureEncoder(json.JSONEncoder):
    """
    Encoder for the Procedure model.
    """

    def default(self, o):
        if callable(o):
            return {
                'type': 'native_function',
                'module': o.__module__,
                'source': inspect.getsourcefile(o),
                'args': [
                    name
                    for name, arg in inspect.signature(o).parameters.items()
                    if arg.kind != arg.VAR_KEYWORD
                ],
            }

        return super().default(o)



from base64 import b64decode
import json
from typing import Optional

from platypush.common.cmd_stream import redis_topic

from . import WSRoute, logger


class WSCommandOutput(WSRoute):
    """
    Websocket route that pushes the output of an executed command to the client
    as it is generated. Mapped to ``/ws/shell``.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, subscriptions=[redis_topic], **kwargs)
        self._id = None

    @classmethod
    def app_name(cls) -> str:
        return 'shell'

    @classmethod
    def path(cls) -> str:
        return f'/ws/{cls.app_name()}'

    def _parse_msg(self, msg: bytes) -> Optional[bytes]:
        parsed_msg = json.loads(msg)
        cmd_id = parsed_msg.get('id')
        output = parsed_msg.get('output')
        if output is None:  # End-of-stream
            raise StopIteration()

        if cmd_id != self._id:
            return None

        return b64decode(output)

    def open(self, *args, **kwargs):
        self._id = next(iter(self.request.arguments['id']), b'').decode() or None
        super().open(*args, **kwargs)

    def run(self) -> None:
        super().run()
        for msg in self.listen():
            try:
                output = self._parse_msg(msg.data)
                if output is None:
                    continue

                self.send(output)
            except StopIteration:
                break
            except Exception as e:
                logger.warning('Failed to parse message: %s', e)
                logger.exception(e)
                continue

        self._io_loop.add_callback(self._ws_close)

    def _ws_close(self):
        if not self.ws_connection:
            return

        self.ws_connection.close(1000, 'Command terminated')



import signal
import subprocess
from typing import Iterable, Optional

from PIL import Image
from PIL.Image import Image as ImageType

from platypush.plugins.camera import CameraPlugin
from platypush.plugins.camera.ffmpeg.model import FFmpegCamera, FFmpegCameraInfo
from platypush.plugins.camera.model.camera import Camera


class CameraFfmpegPlugin(CameraPlugin):
    """
    Plugin to interact with a camera over FFmpeg.
    """

    _camera_class = FFmpegCamera
    _camera_info_class = FFmpegCameraInfo

    def __init__(
        self,
        device: Optional[str] = '/dev/video0',
        input_format: str = 'v4l2',
        ffmpeg_args: Iterable[str] = (),
        **opts,
    ):
        """
        :param device: Path to the camera device (default: ``/dev/video0``).
        :param input_format: FFmpeg input format for the camera device (default: ``v4l2``).
        :param ffmpeg_args: Extra options to be passed to the FFmpeg executable.
        :param opts: Camera options - see constructor of :class:`platypush.plugins.camera.CameraPlugin`.
        """
        super().__init__(device=device, input_format=input_format, **opts)
        self.camera_info.ffmpeg_args = ffmpeg_args or ()  # type: ignore

    def prepare_device(self, device: Camera, **_) -> subprocess.Popen:
        assert isinstance(device, FFmpegCamera)
        warmup_seconds = self._get_warmup_seconds(device)
        ffmpeg = [
            device.info.ffmpeg_bin,
            '-y',
            '-f',
            device.info.input_format,
            '-i',
            device.info.device,
            '-s',
            *(
                (f'{device.info.resolution[0]}x{device.info.resolution[1]}',)
                if device.info.resolution
                else ()
            ),
            '-ss',
            str(warmup_seconds),
            *(('-r', str(device.info.fps)) if device.info.fps else ()),
            '-pix_fmt',
            'rgb24',
            '-f',
            'rawvideo',
            *device.info.ffmpeg_args,
            '-',
        ]

        self.logger.info('Running FFmpeg command: "%s"', ' '.join(ffmpeg))
        proc = subprocess.Popen(ffmpeg, stdout=subprocess.PIPE)
        # Start in suspended mode
        proc.send_signal(signal.SIGSTOP)
        return proc

    def start_camera(self, camera: Camera, *args, preview: bool = False, **kwargs):
        assert isinstance(camera, FFmpegCamera)
        super().start_camera(*args, camera=camera, preview=preview, **kwargs)
        if camera.object:
            camera.object.send_signal(signal.SIGCONT)

    def release_device(self, device: Camera):
        assert isinstance(device, FFmpegCamera)
        if device.object:
            device.object.terminate()
            if device.object.stdout:
                device.object.stdout.close()
            device.object = None  # type: ignore

    def wait_capture(self, camera: Camera):
        assert isinstance(camera, FFmpegCamera)
        if camera.object and camera.object.poll() is None:
            try:
                camera.object.wait(timeout=camera.info.capture_timeout)
            except Exception as e:
                self.logger.warning('Error on FFmpeg capture wait: %s', e)

    def capture_frame(self, device: Camera, *_, **__) -> Optional[ImageType]:
        assert isinstance(device, FFmpegCamera)
        assert device.info.resolution, 'Resolution not set'
        assert device.object.stdout, 'Camera not started'

        raw_size = device.info.resolution[0] * device.info.resolution[1] * 3
        data = device.object.stdout.read(raw_size)
        if len(data) < raw_size:
            return
        return Image.frombytes('RGB', device.info.resolution, data)


# vim:sw=4:ts=4:et:



from contextlib import contextmanager
from dataclasses import dataclass

from sqlalchemy import __version__

sa_version = tuple(map(int, __version__.split('.')))

if sa_version >= (1, 4, 0):
    from sqlalchemy.orm import declarative_base
else:
    from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


@dataclass
class DbContext:
    """
    Context flags for the database session.
    """

    override_definitions: bool = False


_ctx = DbContext()


@contextmanager
def override_definitions():
    """
    Temporarily override the definitions of the entities in the entities
    registry.

    This is useful when the entities are being imported off-context, like
    e.g. in the `inspect` or `alembic` modules.
    """
    _ctx.override_definitions = True
    yield
    _ctx.override_definitions = False


def is_defined(table_name: str) -> bool:
    """
    Check if the given entity class is defined in the entities registry.

    :param table_name: Name of the table associated to the entity class.
    """
    return not _ctx.override_definitions and table_name in Base.metadata



"""Migrate variable table

Revision ID: c39ac404119b
Revises: d030953a871d
Create Date: 2023-04-28 22:35:28.307954

"""

import sqlalchemy as sa
from alembic import op

from platypush.entities import Entity

# revision identifiers, used by Alembic.
revision = 'c39ac404119b'
down_revision = 'd030953a871d'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Get the connection and the existing `variable` table
    conn = op.get_bind()
    metadata = sa.MetaData()
    metadata.reflect(bind=conn)
    VariableOld = metadata.tables.get('variable')

    if VariableOld is None:
        print('The table `variable` does not exist, skipping migration')
        return

    # Create the `variable_new` table
    VariableNew = op.create_table(
        'variable_new',
        sa.Column(
            'id',
            sa.Integer,
            sa.ForeignKey(Entity.id, ondelete='CASCADE'),
            primary_key=True,
        ),
        sa.Column('value', sa.String),
    )

    assert VariableNew is not None, 'Could not create table "variable_new"'

    # Select all existing variables
    existing_vars = {
        var.name: var.value for var in conn.execute(sa.select(VariableOld)).all()
    }

    # Insert all the existing variables as entities
    if existing_vars:
        conn.execute(
            sa.insert(Entity).values(
                [
                    {
                        'external_id': name,
                        'name': name,
                        'type': 'variable',
                        'plugin': 'variable',
                    }
                    for name in existing_vars
                ]
            )
        )

        # Fetch all the newly inserted variables
        new_vars = {
            entity.id: entity.name
            for entity in conn.execute(
                sa.select(Entity.id, Entity.name).where(
                    sa.or_(
                        *[
                            sa.and_(
                                Entity.external_id == name,
                                Entity.type == 'variable',
                                Entity.plugin == 'variable',
                            )
                            for name in existing_vars
                        ]
                    )
                )
            ).all()
        }

        # Insert the mapping on the `variable_new` table
        op.bulk_insert(
            VariableNew,
            [
                {
                    'id': id,
                    'value': existing_vars.get(name),
                }
                for id, name in new_vars.items()
            ],
        )

    # Rename/drop the tables
    op.rename_table('variable', 'variable_old')
    op.rename_table('variable_new', 'variable')
    op.drop_table('variable_old')


def downgrade() -> None:
    # Get the connection and the existing `variable` table
    conn = op.get_bind()
    metadata = sa.MetaData()
    metadata.reflect(bind=conn)
    VariableNew = metadata.tables['variable']

    if VariableNew is None:
        print('The table `variable` does not exist, skipping migration')
        return

    # Create the `variable_old` table
    VariableOld = op.create_table(
        'variable_old',
        sa.Column('name', sa.String, primary_key=True, nullable=False),
        sa.Column('value', sa.String),
    )

    assert VariableOld is not None, 'Could not create table "variable_old"'

    # Select all existing variables
    existing_vars = {
        var.name: var.value
        for var in conn.execute(
            sa.select(Entity.name, VariableNew.c.value).join(
                Entity, Entity.id == VariableNew.c.id
            )
        ).all()
    }

    # Insert the mapping on the `variable_old` table
    if existing_vars:
        op.bulk_insert(
            VariableOld,
            [
                {
                    'name': name,
                    'value': value,
                }
                for name, value in existing_vars.items()
            ],
        )

    # Delete existing references on the `entity` table
    conn.execute(sa.delete(Entity).where(Entity.type == 'variable'))

    # Rename/drop the tables
    op.rename_table('variable', 'variable_new')
    op.rename_table('variable_old', 'variable')
    op.drop_table('variable_new')



from marshmallow import fields, EXCLUDE, INCLUDE
from marshmallow.schema import Schema

from platypush.schemas import DateTime, StrippedString


class TorrentResultSchema(Schema):
    """
    Schema for results returned by
    :meth:`platypush.plugins.torrent.TorrentPlugin.search`.
    """

    # pylint: disable=too-few-public-methods
    class Meta(Schema.Meta):
        """
        Schema metadata.
        """

        missing = EXCLUDE
        ordered = True
        unknown = INCLUDE

    title = StrippedString(
        required=True,
        metadata={
            'description': 'Title of the torrent',
            'example': '2001: A Space Odyssey',
        },
    )

    file = StrippedString(
        metadata={
            'description': (
                'File name of the torrent, if the resource is a local .torrent file'
            ),
            'example': '/home/user/downloads/2001_a_space_odyssey.torrent',
        },
    )

    url = StrippedString(
        metadata={
            'description': 'URL of the torrent, if the resource is a remote torrent',
            'example': 'magnet:?xt=urn:btih:...',
        },
    )

    provider = StrippedString(
        metadata={
            'description': (
                'Provider of the torrent - e.g. `popcorntime`, `yts` or `torrent-csv`'
            ),
            'example': 'popcorntime',
        },
    )

    type = StrippedString(
        metadata={
            'description': 'Type of the torrent',
            'example': 'movies',
        },
    )

    is_media = fields.Boolean(
        metadata={
            'description': 'True if the torrent is a media file',
            'example': True,
        },
    )

    size = fields.Integer(
        missing=0,
        metadata={
            'description': 'Size of the torrent in bytes',
            'example': 123456789,
        },
    )

    duration = fields.Float(
        metadata={
            'description': 'Duration of the torrent in seconds, if applicable',
            'example': 123.45,
        },
    )

    language = StrippedString(
        metadata={
            'description': 'Language of the torrent',
            'example': 'en',
        },
    )

    seeds = fields.Integer(
        missing=0,
        metadata={
            'description': 'Number of seeders',
            'example': 123,
        },
    )

    peers = fields.Integer(
        missing=0,
        metadata={
            'description': 'Number of peers',
            'example': 123,
        },
    )

    image = StrippedString(
        metadata={
            'description': 'URL of the image associated to the torrent',
            'example': 'https://example.com/image.jpg',
        },
    )

    description = StrippedString(
        metadata={
            'description': 'Description of the torrent',
            'example': 'A description of the torrent',
        },
    )

    imdb_id = StrippedString(
        metadata={
            'description': 'IMDb ID of the torrent, if applicable',
            'example': 'tt0062622',
        },
    )

    tvdb_id = StrippedString(
        metadata={
            'description': 'TVDB ID of the torrent, if applicable',
            'example': '76283',
        },
    )

    year = fields.Integer(
        metadata={
            'description': 'Year of the release of the underlying product, if applicable',
            'example': 1968,
        },
    )

    created_at = DateTime(
        metadata={
            'description': 'Creation date of the torrent',
            'example': '2021-01-01T00:00:00',
        },
    )

    quality = StrippedString(
        metadata={
            'description': 'Quality of the torrent, if video/audio',
            'example': '1080p',
        },
    )

    overview = StrippedString(
        metadata={
            'description': 'Overview of the torrent, if it is a TV show',
            'example': 'Overview of the torrent, if it is a TV show',
        },
    )

    trailer = fields.URL(
        metadata={
            'description': 'URL of the trailer of the torrent, if available',
            'example': 'https://example.com/trailer.mp4',
        },
    )

    genres = fields.List(
        fields.String(),
        metadata={
            'description': 'List of genres associated to the torrent',
            'example': ['Sci-Fi', 'Adventure'],
        },
    )

    rating = fields.Float(
        metadata={
            'description': (
                'Rating of the torrent or the underlying product, as a '
                'percentage between 0 and 100'
            ),
            'example': 86.0,
        },
    )

    critic_rating = fields.Float(
        metadata={
            'description': 'Critic rating, if applicable, as a percentage between 0 and 100',
            'example': 86.0,
        },
    )

    community_rating = fields.Float(
        metadata={
            'description': (
                'Community rating, if applicable, as a percentage between 0 and 100'
            ),
            'example': 86.0,
        },
    )

    votes = fields.Integer(
        metadata={
            'description': 'Number of votes, if applicable',
            'example': 123,
        },
    )

    series = StrippedString(
        metadata={
            'description': 'Title of the TV series, if applicable',
            'example': 'Breaking Bad',
        },
    )

    season = fields.Integer(
        metadata={
            'description': 'Season number of the TV series, if applicable',
            'example': 1,
        },
    )

    episode = fields.Integer(
        metadata={
            'description': 'Episode number of the TV series, if applicable',
            'example': 1,
        },
    )

    num_seasons = fields.Integer(
        metadata={
            'description': 'Number of seasons of the TV series, if applicable',
            'example': 5,
        },
    )

    country = StrippedString(
        metadata={
            'description': 'Country of origin of the torrent or the underlying product',
            'example': 'USA',
        },
    )

    network = StrippedString(
        metadata={
            'description': 'Network of the TV series, if applicable',
            'example': 'AMC',
        },
    )

    series_status = StrippedString(
        metadata={
            'description': 'Status of the TV series, if applicable',
            'example': 'Ended',
        },
    )



from abc import ABC
from typing import Optional
from platypush.message.event import Event


class MediaEvent(Event):
    """Base class for media events"""

    def __init__(self, *args, player=None, plugin=None, status=None, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, status=status, **kwargs)


class MediaPlayRequestEvent(MediaEvent):
    """
    Event triggered when a new media playback request is received
    """

    def __init__(
        self, player=None, plugin=None, resource=None, title=None, *args, **kwargs
    ):
        """
        :param resource: File name or URI of the played video
        :type resource: str
        """

        super().__init__(
            *args,
            player=player,
            plugin=plugin,
            resource=resource,
            title=title,
            **kwargs
        )


class MediaPlayEvent(MediaEvent):
    """
    Event triggered when a new media content is played
    """

    def __init__(
        self, player=None, plugin=None, resource=None, title=None, *args, **kwargs
    ):
        """
        :param resource: File name or URI of the played video
        :type resource: str
        """

        super().__init__(
            *args,
            player=player,
            plugin=plugin,
            resource=resource,
            title=title,
            **kwargs
        )


class MediaStopEvent(MediaEvent):
    """
    Event triggered when a media is stopped
    """

    def __init__(self, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, **kwargs)


class MediaPauseEvent(MediaEvent):
    """
    Event triggered when a media playback is paused
    """

    def __init__(self, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, **kwargs)


class MediaResumeEvent(MediaEvent):
    """
    Event triggered when a media playback is resumed
    """

    def __init__(self, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, **kwargs)


class MediaSeekEvent(MediaEvent):
    """
    Event triggered when the time position in the media changes
    """

    def __init__(self, position, player=None, plugin=None, *args, **kwargs):
        super().__init__(
            *args, player=player, plugin=plugin, position=position, **kwargs
        )


class MediaVolumeChangedEvent(MediaEvent):
    """
    Event triggered when the media volume changes
    """

    def __init__(self, volume, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, volume=volume, **kwargs)


class MediaMuteChangedEvent(MediaEvent):
    """
    Event triggered when the media is muted/unmuted
    """

    def __init__(self, mute, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, mute=mute, **kwargs)


class NewPlayingMediaEvent(MediaEvent):
    """
    Event triggered when a new media source is being played
    """

    def __init__(self, player=None, plugin=None, resource=None, *args, **kwargs):
        """
        :param resource: File name or URI of the played resource
        :type resource: str
        """

        super().__init__(
            *args, player=player, plugin=plugin, resource=resource, **kwargs
        )


class MediaDownloadEvent(MediaEvent, ABC):
    """
    Base class for media download events.
    """

    def __init__(
        self,
        *args,
        plugin: str,
        resource: str,
        state: str,
        path: str,
        player: Optional[str] = None,
        size: Optional[int] = None,
        timeout: Optional[int] = None,
        progress: Optional[float] = None,
        started_at: Optional[float] = None,
        ended_at: Optional[float] = None,
        **kwargs
    ):
        """
        :param resource: File name or URI of the downloaded resource
        :param url: Alias for resource
        :param path: Path where the resource is downloaded
        :param state: Download state
        :param size: Size of the downloaded resource in bytes
        :param timeout: Download timeout in seconds
        :param progress: Download progress in percentage, between 0 and 100
        :param started_at: Download start time
        :param ended_at: Download end time
        """

        kwargs.update(
            {
                "resource": resource,
                "path": path,
                "url": resource,
                "state": state,
                "size": size,
                "timeout": timeout,
                "progress": progress,
                "started_at": started_at,
                "ended_at": ended_at,
            }
        )

        super().__init__(*args, player=player, plugin=plugin, **kwargs)


class MediaDownloadStartedEvent(MediaDownloadEvent):
    """
    Event triggered when a media download is started.
    """


class MediaDownloadProgressEvent(MediaDownloadEvent):
    """
    Event triggered when a media download is in progress.
    """


class MediaDownloadCompletedEvent(MediaDownloadEvent):
    """
    Event triggered when a media download is completed.
    """


class MediaDownloadErrorEvent(MediaDownloadEvent):
    """
    Event triggered when a media download fails.
    """

    def __init__(self, error: str, *args, **kwargs):
        """
        :param error: Error message.
        """
        super().__init__(*args, error=error, **kwargs)


class MediaDownloadPausedEvent(MediaDownloadEvent):
    """
    Event triggered when a media download is paused.
    """


class MediaDownloadResumedEvent(MediaDownloadEvent):
    """
    Event triggered when a media download is resumed.
    """


class MediaDownloadCancelledEvent(MediaDownloadEvent):
    """
    Event triggered when a media download is cancelled.
    """


class MediaDownloadClearEvent(MediaDownloadEvent):
    """
    Event triggered when a download is cleared from the queue.
    """


# vim:sw=4:ts=4:et:



import logging

from threading import Thread

from ._model import Command


class CommandBridge(Thread):
    """
    The command bridge is a thread that listens for commands on a command
    queue, proxies them to the Telegram service and returns the result back to
    the response queue.

    This is required because the Telegram service runs in a separate process -
    a requirement because of the Telegram bot API constraints, which requires
    the asyncio event loop to run in the main thread.
    """

    def __init__(self, service, *_, **__):
        from ._service import TelegramService

        super().__init__(name="telegram-service-bridge")
        self.logger = logging.getLogger("platypush:telegram:bridge")
        self._service: TelegramService = service

    def _exec(self, cmd: Command):
        try:
            result = self._service.exec(
                cmd.cmd, *cmd.args, **cmd.kwargs, timeout=cmd.timeout
            )
        except Exception as e:
            result = e

        self._service.result_queue.put_nowait((cmd, result))

    def run(self):
        super().run()

        while self._service.is_running():
            try:
                cmd = self._service.cmd_queue.get()
            except Exception as e:
                self.logger.warning("Error while reading command queue: %s", e)
                continue

            if cmd is None or cmd.is_end_of_service():
                break

            self._exec(cmd)


class ResultBridge(Thread):
    """
    The result bridge is a thread that listens for results on a result queue and
    proxies them to the response queue of the Telegram service.

    This is required because the Telegram service runs in a separate process -
    a requirement because of the Telegram bot API constraints, which requires
    the asyncio event loop to run in the main thread.
    """

    def __init__(self, plugin, *_, **__):
        from . import TelegramPlugin

        super().__init__(name="telegram-service-result-bridge")
        self.logger = logging.getLogger("platypush:telegram:result-bridge")
        self._plugin: TelegramPlugin = plugin

    def run(self):
        super().run()

        while not self._plugin.should_stop():
            try:
                ret = self._plugin.result_queue.get()
            except Exception as e:
                self.logger.warning("Error while reading result queue: %s", e)
                continue

            if not ret:
                break

            cmd, result = ret
            if cmd is None or cmd.is_end_of_service():
                break

            q = self._plugin.response_queues.get(cmd.id)
            if q:
                q.put_nowait(result)


# vim:sw=4:ts=4:et:



from ._async import XmppAsyncMixin
from ._base import XmppBaseMixin
from ._config import XmppConfigMixin
from ._events import XmppEventMixin
from ._state import XmppStateMixin
from ._event_state import XmppEventStateMixin


__all__ = [
    "XmppAsyncMixin",
    "XmppBaseMixin",
    "XmppConfigMixin",
    "XmppEventMixin",
    "XmppEventStateMixin",
    "XmppStateMixin",
]



from abc import ABC
from typing import Iterable, Optional, Union

from platypush.message.event import Event


class XmppEvent(Event, ABC):
    """
    Base class for XMPP events.
    """

    def __init__(self, *args, client_jabber_id: str, **kwargs):
        """
        :param client_jabber_id: The Jabber ID associated to the client connection.
        """
        super().__init__(*args, client_jabber_id=client_jabber_id, **kwargs)


class XmppUserEvent(XmppEvent, ABC):
    """
    Base class for XMPP user events.
    """

    def __init__(self, *args, user_id: str, jid: Optional[str] = None, **kwargs):
        """
        :param user_id: User ID.
        :param jid: The full Jabber ID of the user, if the visibility of the
            full ID including the client identifier is available.
        """
        jid = jid or user_id
        super().__init__(*args, user_id=user_id, jid=jid, **kwargs)


class XmppRoomEvent(XmppEvent, ABC):
    """
    Base class for XMPP room events.
    """

    def __init__(self, *args, room_id: str, **kwargs):
        """
        :param room_id: Room ID.
        """
        super().__init__(*args, room_id=room_id, **kwargs)


class XmppConversationEvent(XmppEvent, ABC):
    """
    Base class for XMPP p2p conversation events.
    """

    def __init__(self, *args, conversation_id: str, **kwargs):
        """
        :param conversation_id: Conversation ID.
        """
        super().__init__(*args, conversation_id=conversation_id, **kwargs)


class XmppRoomOccupantEvent(XmppRoomEvent, XmppUserEvent, ABC):
    """
    Base class for XMPP events about room members.
    """

    def __init__(self, *args, is_self: bool, **kwargs):
        """
        :param is_self: True if the event is about the current user.
        """
        super().__init__(*args, is_self=is_self, **kwargs)


class XmppConversationMemberEvent(XmppConversationEvent, XmppUserEvent, ABC):
    """
    Base class for XMPP events about conversation members.
    """

    def __init__(self, *args, is_self: bool, **kwargs):
        """
        :param is_self: True if the event is about the current user.
        """
        super().__init__(*args, is_self=is_self, **kwargs)


class XmppNickChangedEvent(XmppUserEvent, ABC):
    """
    Base class for XMPP nick changed events.
    """

    def __init__(
        self, *args, old_nick: Optional[str], new_nick: Optional[str], **kwargs
    ):
        """
        :param old_nick: Old nick.
        :param new_nick: New nick.
        """
        super().__init__(*args, old_nick=old_nick, new_nick=new_nick, **kwargs)


class XmppConnectedEvent(XmppEvent):
    """
    Event triggered when the registered XMPP client connects to the server.
    """


class XmppDisconnectedEvent(XmppEvent):
    """
    Event triggered when the registered XMPP client disconnects from the server.
    """

    def __init__(self, *args, reason: Optional[Union[str, Exception]] = None, **kwargs):
        """
        :param reason: The reason of the disconnection.
        """
        super().__init__(*args, reason=str(reason) if reason else None, **kwargs)


class XmppUserAvailableEvent(XmppUserEvent):
    """
    Event triggered when a user the client is subscribed to becomes available.
    """


class XmppUserUnavailableEvent(XmppUserEvent):
    """
    Event triggered when a user the client is subscribed to becomes unavailable.
    """


class XmppRoomUserAvailableEvent(XmppRoomOccupantEvent):
    """
    Event triggered when a user in a joined room becomes available.
    """


class XmppRoomUserUnavailableEvent(XmppRoomOccupantEvent):
    """
    Event triggered when a user in a joined room becomes unavailable.
    """


class XmppMessageReceivedEvent(XmppUserEvent):
    """
    Event triggered when the registered XMPP client receives a message.
    """

    def __init__(self, *args, body: str, **kwargs):
        """
        :param body: The body of the message.
        """
        super().__init__(*args, body=body, **kwargs)


class XmppRoomMessageReceivedEvent(XmppMessageReceivedEvent, XmppRoomOccupantEvent):
    """
    Event triggered when a message is received on a multi-user conversation
    joined by the client.
    """


class XmppRoomInviteAcceptedEvent(XmppRoomEvent):
    """
    Event triggered when an invite to a room is accepted.
    """


class XmppRoomInviteRejectedEvent(XmppRoomEvent):
    """
    Event triggered when an invite to a room is rejected.
    """


class XmppRoomJoinEvent(XmppRoomOccupantEvent):
    """
    Event triggered when a user joins a room.
    """

    def __init__(self, *args, members: Optional[Iterable[str]] = None, **kwargs):
        """
        :param members: List of IDs of the joined members.
        """
        super().__init__(*args, members=list(set(members or [])), **kwargs)


class XmppRoomLeaveEvent(XmppRoomOccupantEvent):
    """
    Event triggered when a user leaves a room.
    """


class XmppRoomEnterEvent(XmppRoomOccupantEvent):
    """
    Event triggered when a user first enters a room.
    """


class XmppRoomExitEvent(XmppRoomOccupantEvent):
    """
    Event triggered when a user exits a room.
    """

    def __init__(self, *args, reason: Optional[str] = None, **kwargs):
        """
        :param reason: Exit reason.
        """
        super().__init__(*args, reason=reason, **kwargs)


class XmppRoomTopicChangedEvent(XmppRoomEvent):
    """
    Event triggered when the topic of a room is changed.
    """

    def __init__(
        self,
        *args,
        topic: Optional[str] = None,
        changed_by: Optional[str] = None,
        **kwargs
    ):
        """
        :param topic: New room topic.
        :param changed_by: Nick of the user who changed the topic.
        """
        super().__init__(*args, topic=topic, changed_by=changed_by, **kwargs)


class XmppPresenceChangedEvent(XmppUserEvent):
    """
    Event triggered when the reported presence of a user in the contacts list
    changes.
    """

    def __init__(self, *args, status: Optional[str], **kwargs):
        """
        :param status: New presence status.
        """
        super().__init__(*args, status=status, **kwargs)


class XmppRoomPresenceChangedEvent(XmppPresenceChangedEvent, XmppRoomEvent):
    """
    Event triggered when the reported presence of a user in a room changes.
    """


class XmppRoomAffiliationChangedEvent(XmppRoomOccupantEvent):
    """
    Event triggered when the affiliation of a user in a room changes.
    """

    def __init__(
        self,
        *args,
        affiliation: str,
        changed_by: Optional[str] = None,
        reason: Optional[str] = None,
        **kwargs
    ):
        """
        :param affiliation: New affiliation.
        :param changed_by: Nick of the user who changed the affiliation.
        :param reason: Affiliation change reason.
        """
        super().__init__(
            *args,
            affiliation=affiliation,
            changed_by=changed_by,
            reason=reason,
            **kwargs
        )


class XmppRoomRoleChangedEvent(XmppRoomOccupantEvent):
    """
    Event triggered when the role of a user in a room changes.
    """

    def __init__(
        self,
        *args,
        role: str,
        changed_by: Optional[str] = None,
        reason: Optional[str] = None,
        **kwargs
    ):
        """
        :param role: New role.
        :param changed_by: Nick of the user who changed the role.
        :param reason: Role change reason.
        """
        super().__init__(
            *args, role=role, changed_by=changed_by, reason=reason, **kwargs
        )


class XmppRoomNickChangedEvent(XmppNickChangedEvent, XmppRoomOccupantEvent):
    """
    Event triggered when a user in a room changes their nick.
    """


class XmppRoomInviteEvent(XmppRoomEvent, XmppUserEvent):
    """
    Event triggered when the client is invited to join a room.
    """

    def __init__(
        self,
        *args,
        mode: str,
        password: Optional[str] = None,
        reason: Optional[str] = None,
        **kwargs
    ):
        """
        :param user_id: The user who sent the invite.
        :param mode: Invite mode, either ``DIRECT`` or ``MEDIATED``.
        :param password: The room password.
        :param reason: Optional invite reason.
        """
        super().__init__(*args, mode=mode, password=password, reason=reason, **kwargs)


class XmppConversationAddedEvent(XmppConversationEvent):
    """
    Event triggered when a conversation is added to the client's list.
    """

    def __init__(self, *args, members: Optional[Iterable[str]] = None, **kwargs):
        """
        :param members: Jabber IDs of the conversation members.
        """
        super().__init__(*args, members=list(set(members or [])), **kwargs)


class XmppConversationEnterEvent(XmppConversationEvent):
    """
    Event triggered when the user enters a conversation.
    """


class XmppConversationExitEvent(XmppConversationEvent):
    """
    Event triggered when the user exits a conversation.
    """


class XmppConversationNickChangedEvent(
    XmppNickChangedEvent, XmppConversationMemberEvent
):
    """
    Event triggered when a user in a p2p conversation changes their nick.
    """


class XmppConversationJoinEvent(XmppConversationMemberEvent):
    """
    Event triggered when a user enters a conversation.
    """


class XmppConversationLeaveEvent(XmppConversationMemberEvent):
    """
    Event triggered when the user leaves a conversation.
    """


class XmppContactAddRequestEvent(XmppUserEvent):
    """
    Event triggered when a user adds the client Jabber ID to their contacts
    list.
    """


class XmppContactAddRequestAcceptedEvent(XmppUserEvent):
    """
    Event triggered when a user contact add request is accepted.
    """


class XmppContactAddRequestRejectedEvent(XmppUserEvent):
    """
    Event triggered when a user contact add request is rejected.
    """



from typing import Callable, Optional, Sequence, Union

import pvcheetah

from platypush.message.event.assistant import (
    ConversationTimeoutEvent,
    SpeechRecognizedEvent,
)

from ._base import BaseProcessor


class SttProcessor(BaseProcessor):
    """
    Implementation of the speech-to-text processor using the Picovoice Cheetah
    engine.
    """

    def __init__(
        self, *args, get_cheetah_args: Callable[[], dict] = lambda: {}, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._get_cheetah_args = get_cheetah_args
        # model_path -> Cheetah instance cache
        self._cheetah = {self._model_path: pvcheetah.create(**self._get_cheetah_args())}

    @property
    def _model_path(self) -> Optional[str]:
        return self._get_cheetah_args().get('model_path')

    @property
    def sample_rate(self) -> int:
        return self._get_cheetah().sample_rate

    @property
    def frame_length(self) -> int:
        return self._get_cheetah().frame_length

    def _get_cheetah(self) -> pvcheetah.Cheetah:
        with self._state_lock:
            if not self._cheetah.get(self._model_path):
                self.logger.debug(
                    'Creating Cheetah instance for model %s', self._model_path
                )
                self._cheetah[self._model_path] = pvcheetah.create(
                    **self._get_cheetah_args()
                )
                self.logger.debug(
                    'Cheetah instance created for model %s', self._model_path
                )

            return self._cheetah[self._model_path]

    def process(
        self, audio: Sequence[int]
    ) -> Optional[Union[SpeechRecognizedEvent, ConversationTimeoutEvent]]:
        event = None
        cheetah = self._get_cheetah()
        partial_transcript, self._ctx.is_final = cheetah.process(audio)
        last_transcript = self._ctx.transcript

        # Concatenate the partial transcript to the context
        if partial_transcript:
            self._ctx.transcript += partial_transcript
            self.logger.info(
                'Partial transcript: %s, is_final: %s',
                self._ctx.transcript,
                self._ctx.is_final,
            )

        # If the transcript is final or the conversation timed out, then
        # process and return whatever is available in the context
        if self._ctx.is_final or self._ctx.timed_out:
            phrase = cheetah.flush() or ''
            self._ctx.transcript += phrase
            if self._ctx.transcript and self._ctx.transcript != last_transcript:
                self.logger.debug('Processed STT transcript: %s', self._ctx.transcript)
                last_transcript = self._ctx.transcript

            phrase = self._ctx.transcript
            phrase = (phrase[:1].lower() + phrase[1:]).strip()
            event = (
                SpeechRecognizedEvent(phrase=phrase)
                if phrase
                else ConversationTimeoutEvent()
            )

            self.reset()

        return event

    def reset(self):
        if not self._enabled:
            return

        with self._state_lock:
            super().reset()
            self._get_cheetah().flush()

    def stop(self):
        if not self._enabled:
            return

        super().stop()

        with self._state_lock:
            objs = self._cheetah.copy()
            for key, obj in objs.items():
                obj.delete()
                self._cheetah.pop(key)



from dataclasses import dataclass
import math
from typing import Tuple

from flask import Blueprint, make_response, request


logo = Blueprint('logo', __name__)

# Declare routes list
__routes__ = [
    logo,
]


@dataclass
class Gear:
    """
    A utility class used to model the gears in the application's logo.
    """

    center: Tuple[float, float]
    outer_radius: float
    inner_radius: float
    color: str = "currentColor"
    background: str = ""
    num_spikes: int = 0
    spike_max_base: float = 0
    spike_min_base: float = 0
    spike_height: float = 0
    alpha_offset: float = 0

    def to_svg(self) -> str:
        """
        Returns the SVG representation of a gear.
        """

        # Generate the basic circle
        svg = f"""
        <circle
            cx="{self.center[0]}" cy="{self.center[1]}"
            r="{self.outer_radius - (self.inner_radius / math.pi)}"
            stroke-width="{self.inner_radius}"
            stroke="{self.color}"
            fill="none" />
        """

        # Generate the spikes
        for i in range(self.num_spikes):
            # Iterate for alpha -> [0, 2*pi]
            alpha = (2 * math.pi * i) / self.num_spikes
            # Calculate the base angle for the major base of the gear polygon
            maj_delta_alpha = math.asin(self.spike_max_base / (2 * self.outer_radius))
            # Calculate the points of the gear polygon's major base
            maj_base = (
                (
                    self.center[0]
                    + self.outer_radius
                    * math.cos(alpha + maj_delta_alpha + self.alpha_offset),
                    self.center[1]
                    + self.outer_radius
                    * math.sin(alpha + maj_delta_alpha + self.alpha_offset),
                ),
                (
                    self.center[0]
                    + self.outer_radius
                    * math.cos(alpha - maj_delta_alpha + self.alpha_offset),
                    self.center[1]
                    + self.outer_radius
                    * math.sin(alpha - maj_delta_alpha + self.alpha_offset),
                ),
            )

            # Height of the gear relative to the circle's center
            h = self.outer_radius * math.cos(maj_delta_alpha) + self.spike_height
            # Calculate the base angle for the minor base of the gear polygon
            min_delta_alpha = math.asin(self.spike_min_base / (2 * h))
            # Calculate the points of the gear polygon's minor base
            min_base = (
                (
                    self.center[0]
                    + h * math.cos(alpha - min_delta_alpha + self.alpha_offset),
                    self.center[1]
                    + h * math.sin(alpha - min_delta_alpha + self.alpha_offset),
                ),
                (
                    self.center[0]
                    + h * math.cos(alpha + min_delta_alpha + self.alpha_offset),
                    self.center[1]
                    + h * math.sin(alpha + min_delta_alpha + self.alpha_offset),
                ),
            )

            # Flatten the polygon's points
            svg_points = " ".join(
                [f"{point[0]},{point[1]}" for point in [*maj_base, *min_base]]
            )

            # Serialize the gear polygon to SVG
            svg += f"""
        <polygon points="{svg_points}" stroke="{self.color}" fill="{self.color}" />"""

        return svg


# Properties of the two gears on the logo
gears = [
    Gear(
        center=(32.9, 34.5),
        outer_radius=22.6,
        inner_radius=12.4,
        num_spikes=12,
        spike_max_base=9,
        spike_min_base=4.3,
        spike_height=10.16,
    ),
    Gear(
        center=(65.5, 70.5),
        outer_radius=14.4,
        inner_radius=8.5,
        num_spikes=7,
        spike_max_base=9,
        spike_min_base=4.3,
        spike_height=7.5,
        alpha_offset=math.pi / 6.6,
    ),
]


template_start = """
<svg version="1.1"
     width="{width}" height="{height}"
     viewBox="0 0 100 100"
     preserveAspectRatio="none"
     xmlns="http://www.w3.org/2000/svg">
  <defs>
    <linearGradient id="triangleGradient">
        <stop offset="0%" stop-color="#8acb45" />
        <stop offset="50%" stop-color="#6bbb4c" />
        <stop offset="100%" stop-color="#5cb450" />
    </linearGradient>
  </defs>

  <rect width="100%" height="100%" fill="{bg_color}" />
"""

template_end = "\n</svg>"


@logo.route('/logo.svg', methods=['GET'])
def logo_path():
    """
    This path dynamically generates the logo image as a parametrizable vector SVG.

    Parameters:

        - ``size``: Size of the image in pixels (default: 256)
        - ``bg``: Background color (default: "none")
        - ``fg``: Foreground color (default: "currentColor")

    """
    size = request.args.get("size", 256)
    bg = request.args.get("bg", "none")
    fg = request.args.get("fg", "currentColor")
    svg = template_start.format(
        width=size,
        height=size,
        bg_color=bg,
    )

    for gear in gears:
        gear.color = fg
        gear.background = bg
        svg += gear.to_svg()

    # "Play" triangle on the logo
    svg += """\n\t\t<polygon points="67,47 67,3 99,25.3" fill="url(#triangleGradient)" />"""
    svg += template_end

    rs = make_response(svg)
    rs.headers.update({"Content-Type": "image/svg+xml"})
    return rs


# vim:sw=4:ts=4:et:



from logging import getLogger
from socket import socket

from platypush.commands import Command


# pylint: disable=too-few-public-methods
class CommandWriter:
    """
    Writes command objects to file-like I/O objects.
    """

    def __init__(self):
        self.logger = getLogger(__name__)

    def write(self, cmd: Command, sock: socket):
        """
        Writes a command to a file-like I/O object.

        :param cmd: The command to write.
        :param fp: The file-like I/O object to write to.
        """

        buf = cmd.to_bytes()
        sock.sendall(buf)



import subprocess
import threading
import time

from abc import ABC
from typing import Optional, Tuple

from platypush.plugins.camera.model.camera import Camera
from platypush.plugins.camera.model.writer import (
    VideoWriter,
    FileVideoWriter,
    StreamWriter,
)


class FFmpegWriter(VideoWriter, ABC):
    """
    Generic FFmpeg encoder for camera frames.
    """

    def __init__(
        self,
        *args,
        input_file: str = '-',
        input_format: str = 'rawvideo',
        output_file: str = '-',
        output_format: Optional[str] = None,
        pix_fmt: Optional[str] = None,
        output_opts: Optional[Tuple] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.input_file = input_file
        self.input_format = input_format
        self.output_format = output_format
        self.output_file = output_file
        self.width, self.height = self.camera.effective_resolution()
        self.pix_fmt = pix_fmt
        self.output_opts = output_opts or ()

        self.logger.info('Starting FFmpeg. Command: ' + ' '.join(self.ffmpeg_args))
        self.ffmpeg = subprocess.Popen(
            self.ffmpeg_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )

    @property
    def ffmpeg_args(self):
        return [
            self.camera.info.ffmpeg_bin,
            '-y',
            '-f',
            self.input_format,
            *(('-pix_fmt', self.pix_fmt) if self.pix_fmt else ()),
            '-s',
            f'{self.width}x{self.height}',
            '-r',
            str(self.camera.info.fps),
            '-i',
            self.input_file,
            *(('-f', self.output_format) if self.output_format else ()),
            *self.output_opts,
            *(
                ('-vcodec', self.camera.info.output_codec)
                if self.camera.info.output_codec
                else ()
            ),
            self.output_file,
        ]

    def is_closed(self):
        return self.closed or not self.ffmpeg or self.ffmpeg.poll() is not None

    def write(self, image):
        if self.is_closed():
            return

        try:
            self.ffmpeg.stdin.write(image.convert('RGB').tobytes())
        except Exception as e:
            self.logger.warning('FFmpeg send error: %s', e)
            self.close()

    def close(self):
        if not self.is_closed():
            if self.ffmpeg and self.ffmpeg.stdin:
                try:
                    self.ffmpeg.stdin.close()
                except OSError:
                    pass

            if self.ffmpeg:
                self.ffmpeg.terminate()
                try:
                    self.ffmpeg.wait(timeout=5.0)
                except subprocess.TimeoutExpired:
                    self.logger.warning('FFmpeg has not returned - killing it')
                    self.ffmpeg.kill()

            if self.ffmpeg and self.ffmpeg.stdout:
                try:
                    self.ffmpeg.stdout.close()
                except OSError:
                    pass

        self.ffmpeg = None
        super().close()


class FFmpegFileWriter(FileVideoWriter, FFmpegWriter):
    """
    Write camera frames to a file using FFmpeg.
    """

    def __init__(self, *args, output_file: str, **kwargs):
        super().__init__(*args, output_file=output_file, pix_fmt='rgb24', **kwargs)


class FFmpegStreamWriter(StreamWriter, FFmpegWriter, ABC):
    """
    Stream camera frames using FFmpeg.
    """

    def __init__(
        self, *args, output_format: str, output_opts: Optional[Tuple] = None, **kwargs
    ):
        super().__init__(
            *args,
            pix_fmt='rgb24',
            output_format=output_format,
            output_opts=output_opts
            or (
                '-tune',
                'zerolatency',
                '-preset',
                'superfast',
                '-trellis',
                '0',
                '-fflags',
                'nobuffer',
            ),
            **kwargs,
        )
        self._reader = threading.Thread(target=self._reader_thread)
        self._reader.start()

    def encode(self, image) -> bytes:
        return image.convert('RGB').tobytes()

    def _reader_thread(self):
        start_time = time.time()

        while not self.is_closed():
            try:
                data = self.ffmpeg.stdout.read(1 << 15)
            except Exception as e:
                self.logger.warning('FFmpeg reader error: %s', e)
                break

            if not data:
                continue

            if self.frame is None:
                latency = time.time() - start_time
                self.logger.info('FFmpeg stream latency: %d secs', latency)

            with self.ready:
                self.frame = data
                self.frame_time = time.time()
                self.ready.notify_all()

            self._sock_send(self.frame)

    def write(self, image):
        if self.is_closed():
            return

        data = self.encode(image)
        try:
            self.ffmpeg.stdin.write(data)
        except Exception as e:
            self.logger.warning('FFmpeg send error: %s', e)
            self.close()

    def close(self):
        super().close()
        if (
            self._reader
            and self._reader.is_alive()
            and threading.get_ident() != self._reader.ident
        ):
            self._reader.join(timeout=5.0)
            self._reader = None


class MKVStreamWriter(FFmpegStreamWriter):
    mimetype = 'video/webm'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, output_format='matroska', **kwargs)


class H264StreamWriter(FFmpegStreamWriter):
    mimetype = 'video/h264'

    def __init__(self, *args, camera: Camera, **kwargs):
        if not camera.info.output_codec:
            camera.info.output_codec = 'libxvid'
        super().__init__(*args, camera=camera, output_format='h264', **kwargs)


class H265StreamWriter(FFmpegStreamWriter):
    mimetype = 'video/h265'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, output_format='h265', **kwargs)


# vim:sw=4:ts=4:et:



from marshmallow import fields, Schema, INCLUDE


class HidDeviceSchema(Schema):
    class Meta:
        unknown = INCLUDE

    path = fields.String(
        metadata={
            'description': 'Path to the raw HID device',
            'example': '/dev/hidraw0',
        },
    )

    serial_number = fields.String(
        metadata={
            'description': 'Serial number',
            'example': '00:11:22:33:44:55',
        },
    )

    vendor_id = fields.Integer(
        metadata={
            'description': 'Vendor ID',
            'example': 1234,
        },
    )

    product_id = fields.Integer(
        metadata={
            'description': 'Product ID',
            'example': 4321,
        },
    )

    manufacturer_string = fields.String(
        metadata={
            'description': 'Manufacturer custom string',
            'example': 'foo',
        },
    )

    product_string = fields.String(
        metadata={
            'description': 'Main name of the product',
            'example': 'My Device',
        },
    )


class HidMonitoredDeviceSchema(HidDeviceSchema):
    notify_only_if_changed = fields.Boolean(
        load_default=True,
        metadata={
            'description': 'If set to true (default), only changes in the '
            'values of the device will trigger events. So if you are e.g. '
            'monitoring the state of a joystick, only changes in the pressed '
            'buttons will trigger events.',
        },
    )

    data_size = fields.Integer(
        load_default=64,
        metadata={
            'description': 'How many bytes should be read from the device on '
            'each iteration (default: 64)',
        },
    )

    poll_seconds = fields.Float(
        load_default=0,
        metadata={
            'description': 'How often we should wait before data reads '
            '(default: no wait)'
        },
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for attr in (
            'path',
            'serial_number',
            'vendor_id',
            'product_id',
            'manufacturer_string',
            'product_string',
        ):
            self._declared_fields[attr].metadata['description'] += ' (optional)'



import ast
import contextlib
import datetime
import functools
import hashlib
import importlib
import inspect
import logging
import os
import pathlib
import re
import signal
import socket
import ssl
import time
import urllib.request
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec
from multiprocessing import Lock as PLock
from tempfile import gettempdir
from threading import Event, Lock as TLock
from typing import Generator, Optional, Tuple, Type, Union

from dateutil import parser, tz
from redis import ConnectionPool, Redis
from rsa.key import PublicKey, PrivateKey, newkeys

logger = logging.getLogger('utils')
Lock = Union[PLock, TLock]  # type: ignore

redis_pools: dict[Tuple[str, int], ConnectionPool] = {}


def get_module_and_method_from_action(action):
    """
    Input: action=music.mpd.play
    Output: ('music.mpd', 'play')
    """

    tokens = action.split('.')
    module_name = str.join('.', tokens[:-1])
    method_name = tokens[-1:][0]
    return module_name, method_name


def get_message_class_by_type(msgtype):
    """Gets the class of a message type given as string"""

    try:
        module = importlib.import_module('platypush.message.' + msgtype)
    except ImportError as e:
        logger.warning('Unsupported message type %s', msgtype)
        raise RuntimeError(e) from e

    cls_name = msgtype[0].upper() + msgtype[1:]

    try:
        msgclass = getattr(module, cls_name)
    except AttributeError as e:
        logger.warning('No such class in %s: %s', module.__name__, cls_name)
        raise RuntimeError(e) from e

    return msgclass


def get_event_class_by_type(type):  # pylint: disable=redefined-builtin
    """Gets an event class by type name"""
    event_module = importlib.import_module('.'.join(type.split('.')[:-1]))
    return getattr(event_module, type.split('.')[-1])


def get_plugin_module_by_name(plugin_name):
    """Gets the module of a plugin by name (e.g. "music.mpd" or "media.vlc")"""

    module_name = 'platypush.plugins.' + plugin_name
    try:
        return importlib.import_module('platypush.plugins.' + plugin_name)
    except ImportError as e:
        logger.error('Cannot import %s: %s', module_name, e)
        return None


def get_backend_module_by_name(backend_name):
    """Gets the module of a backend by name (e.g. "backend.http" or "backend.mqtt")"""

    module_name = 'platypush.backend.' + backend_name
    try:
        return importlib.import_module('platypush.backend.' + backend_name)
    except ImportError as e:
        logger.error('Cannot import %s: %s', module_name, e)
        return None


def get_plugin_class_by_name(plugin_name) -> Optional[type]:
    """Gets the class of a plugin by name (e.g. "music.mpd" or "media.vlc")"""

    module = get_plugin_module_by_name(plugin_name)
    if not module:
        return None

    class_name = getattr(
        module, ''.join([_.capitalize() for _ in plugin_name.split('.')]) + 'Plugin'
    )
    try:
        return getattr(
            module, ''.join([_.capitalize() for _ in plugin_name.split('.')]) + 'Plugin'
        )
    except Exception as e:
        logger.error('Cannot import class %s: %s', class_name, e)
        return None


def get_plugin_name_by_class(plugin) -> str:
    """Gets the common name of a plugin (e.g. "music.mpd" or "media.vlc") given its class."""

    from platypush.plugins import Plugin

    if isinstance(plugin, Plugin):
        plugin = plugin.__class__

    class_name = plugin.__name__
    class_tokens = [
        token.lower()
        for token in re.sub(r'([A-Z])', r' \1', class_name).split(' ')
        if token.strip() and token != 'Plugin'
    ]

    return '.'.join(class_tokens)


def get_backend_class_by_name(backend_name: str) -> Optional[type]:
    """Gets the class of a backend by name (e.g. "backend.http" or "backend.mqtt")"""

    module = get_backend_module_by_name(backend_name)
    if not module:
        return None

    class_name = getattr(
        module,
        ''.join(
            [
                token.capitalize()
                for i, token in enumerate(backend_name.split('.'))
                if not (i == 0 and token == 'backend')
            ]
        )
        + 'Backend',
    )
    try:
        return getattr(
            module,
            ''.join([_.capitalize() for _ in backend_name.split('.')]) + 'Backend',
        )
    except Exception as e:
        logger.error('Cannot import class %s: %s', class_name, e)
        return None


def get_backend_name_by_class(backend) -> str:
    """Gets the common name of a backend (e.g. "http" or "mqtt") given its class."""

    from platypush.backend import Backend

    if isinstance(backend, Backend):
        backend = backend.__class__

    class_name = backend.__name__
    class_tokens = [
        token.lower()
        for token in re.sub(r'([A-Z])', r' \1', class_name).split(' ')
        if token.strip() and token != 'Backend'
    ]

    return '.'.join(class_tokens)


def set_timeout(seconds, on_timeout):
    """
    Set a function to be called if timeout expires without being cleared.
    It only works on the main thread.

    Params:
        seconds    -- Timeout in seconds
        on_timeout -- Function invoked on timeout unless clear_timeout is called before
    """

    def _sighandler(*_):
        on_timeout()

    signal.signal(signal.SIGALRM, _sighandler)
    signal.alarm(seconds)


def clear_timeout():
    """Clear any previously set timeout"""
    signal.alarm(0)


def get_hash(s):
    """Get the SHA256 hash hex digest of a string input"""
    return hashlib.sha256(s.encode('utf-8')).hexdigest()


def get_decorators(cls, climb_class_hierarchy=False):
    """
    Get the decorators of a class as a {"decorator_name": [list of methods]} dictionary

    :param cls: Class type
    :param climb_class_hierarchy: If set to True (default: False), it will search return the decorators in the parent
        classes as well
    :type climb_class_hierarchy: bool
    """

    decorators = {}

    def visit_FunctionDef(node):
        for n in node.decorator_list:
            if isinstance(n, ast.Call):
                name = (
                    n.func.attr
                    if isinstance(n.func, ast.Attribute)
                    else n.func.id  # type: ignore
                )
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[name] = decorators.get(name, set())
            decorators[name].add(node.name)

    if climb_class_hierarchy:
        targets = inspect.getmro(cls)
    else:
        targets = [cls]

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef

    for target in targets:
        with contextlib.suppress(TypeError):
            node_iter.visit(ast.parse(inspect.getsource(target)))

    return decorators


def get_redis_queue_name_by_message(msg) -> Optional[str]:
    """
    Get the Redis queue name for the response(s) associated to a request
    message.

    :param msg: Input message, as a :class:`platypush.message.request.Request`
        object.
    """
    from platypush.message.request import Request

    if not isinstance(msg, Request):
        logger.warning('Not a valid request (type: %s): %s', type(msg), msg)
        return None
    return f'platypush/responses/{msg.id}' if msg.id else None


def _get_ssl_context(
    context_type=None, ssl_cert=None, ssl_key=None, ssl_cafile=None, ssl_capath=None
):
    if not context_type:
        ssl_context = ssl.create_default_context(cafile=ssl_cafile, capath=ssl_capath)
    else:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    assert ssl_cert, 'No certificate specified'
    if ssl_cafile or ssl_capath:
        ssl_context.load_verify_locations(cafile=ssl_cafile, capath=ssl_capath)

    ssl_context.load_cert_chain(
        certfile=os.path.abspath(os.path.expanduser(ssl_cert)),
        keyfile=os.path.abspath(os.path.expanduser(ssl_key)) if ssl_key else None,
    )

    return ssl_context


def get_ssl_context(ssl_cert=None, ssl_key=None, ssl_cafile=None, ssl_capath=None):
    """
    Generic builder for SSL context.
    """
    return _get_ssl_context(
        context_type=None,
        ssl_cert=ssl_cert,
        ssl_key=ssl_key,
        ssl_cafile=ssl_cafile,
        ssl_capath=ssl_capath,
    )


def get_ssl_server_context(
    ssl_cert=None, ssl_key=None, ssl_cafile=None, ssl_capath=None
):
    """
    Builder for a server-side SSL context.
    """
    return _get_ssl_context(
        context_type=ssl.PROTOCOL_TLS_SERVER,
        ssl_cert=ssl_cert,
        ssl_key=ssl_key,
        ssl_cafile=ssl_cafile,
        ssl_capath=ssl_capath,
    )


def get_ssl_client_context(
    ssl_cert=None, ssl_key=None, ssl_cafile=None, ssl_capath=None
):
    """
    Builder for a client-side SSL context.
    """
    return _get_ssl_context(
        context_type=ssl.PROTOCOL_TLS_CLIENT,
        ssl_cert=ssl_cert,
        ssl_key=ssl_key,
        ssl_cafile=ssl_cafile,
        ssl_capath=ssl_capath,
    )


def find_bins_in_path(bin_name):
    """
    Search for a binary in the PATH variable.
    """
    return [
        os.path.join(p, bin_name)
        for p in os.environ.get('PATH', '').split(':')
        if os.path.isfile(os.path.join(p, bin_name))
        and (os.name == 'nt' or os.access(os.path.join(p, bin_name), os.X_OK))
    ]


def find_files_by_ext(directory, *exts):
    """
    Finds all the files in the given directory with the provided extensions.
    """

    if not exts:
        raise AttributeError('No extensions provided')

    if not os.path.isdir(directory):
        raise AttributeError(f'{directory} is not a valid directory')

    min_len = len(min(exts, key=len))
    max_len = len(max(exts, key=len))
    result = []

    for _, __, files in os.walk(directory):
        for i in range(min_len, max_len + 1):
            result += [f for f in files if f[-i:] in exts]

    return result


def is_process_alive(pid: int) -> bool:
    """
    :param pid: Process ID.
    :return: True if the process with the given PID is alive.
    """
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def get_ip_or_hostname() -> str:
    """
    Get the the default IP address or hostname of the machine.
    """
    ip = socket.gethostbyname(socket.gethostname())
    if ip.startswith('127.') or ip.startswith('::1'):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect(('10.255.255.255', 1))
            ip = sock.getsockname()[0]
            sock.close()
        except Exception as e:
            logger.debug(e)

    return ip


def get_mime_type(resource: str) -> Optional[str]:
    """
    Get the MIME type of the given resource.

    :param resource: The resource to get the MIME type for - it can be a file
        path or a URL.
    """
    import magic

    if resource.startswith('file://'):
        offset = len('file://')
        resource = resource[offset:]

    if resource.startswith('http://') or resource.startswith('https://'):
        with urllib.request.urlopen(resource) as response:
            return response.info().get_content_type()
    else:
        if hasattr(magic, 'detect_from_filename'):
            mime = magic.detect_from_filename(resource)  # type: ignore
        elif hasattr(magic, 'from_file'):
            mime = magic.from_file(resource, mime=True)
        else:
            raise RuntimeError(
                'The installed magic version provides neither detect_from_filename nor from_file'
            )

        if mime:
            return mime.mime_type if hasattr(mime, 'mime_type') else mime  # type: ignore

    return None


def camel_case_to_snake_case(string):
    """
    Utility function to convert CamelCase to snake_case.
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def grouper(n, iterable, fillvalue=None):
    """
    Split an iterable in groups of max N elements.
    grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    """
    from itertools import zip_longest

    args = [iter(iterable)] * n

    if fillvalue:
        return zip_longest(*args, fillvalue=fillvalue)

    for chunk in zip_longest(*args):
        yield filter(None, chunk)

    return


def is_functional_procedure(obj) -> bool:
    """
    Check if the given object is a functional procedure.
    """
    return callable(obj) and hasattr(obj, 'procedure')


def is_functional_hook(obj) -> bool:
    """
    Check if the given object is a functional hook.
    """
    return callable(obj) and hasattr(obj, 'hook')


def is_functional_cron(obj) -> bool:
    """
    Check if the given object is a functional cron.
    """
    return callable(obj) and hasattr(obj, 'cron') and hasattr(obj, 'cron_expression')


def run(action, *args, **kwargs):
    """
    Run the given action with the given arguments. Example:

        >>> from platypush.utils import run
        >>> run('music.mpd.play', resource='file:///home/user/music.mp3')

    """
    from platypush.config import Config
    from platypush.context import get_plugin
    from platypush.procedure import Procedure

    if action.startswith('procedure.'):
        procedure_name = action.removeprefix('procedure.')
        procedures = Config.get_procedures()
        procedure = procedures.get(procedure_name)
        if not procedure:
            raise RuntimeError(f'No such procedure: {procedure_name}')

        if isinstance(procedure, dict):
            procedure = Procedure.build(
                name=procedure_name,
                requests=procedure.get('actions', []),
                args=procedure.get('args', {}),
                _async=procedure.get('async', False),
            )

            return procedure.execute(*args, **kwargs)

        return procedure(*args, **kwargs)

    (module_name, method_name) = get_module_and_method_from_action(action)
    plugin = get_plugin(module_name)
    method = getattr(plugin, method_name)
    response = method(*args, **kwargs)

    if response.errors:
        raise RuntimeError(response.errors[0])

    return response.output


def generate_rsa_key_pair(
    key_file: Optional[str] = None, size: int = 2048
) -> Tuple[PublicKey, PrivateKey]:
    """
    Generate an RSA key pair.

    :param key_file: Target file for the private key (the associated public key will be stored in ``<key_file>.pub``.
        If no key file is specified then the public and private keys will be returned in ASCII format in a dictionary
        with the following structure:

            .. code-block:: json

                {
                    "private": "private key here",
                    "public": "public key here"
                }

    :param size: Key size (default: 2048 bits).
    :return: A tuple with the generated ``(priv_key_str, pub_key_str)``.
    """
    logger.info('Generating RSA keypair')
    pub_key, priv_key = newkeys(size)
    logger.info('Generated RSA keypair')
    public_key_str = pub_key.save_pkcs1('PEM').decode()
    private_key_str = priv_key.save_pkcs1('PEM').decode()

    if key_file:
        logger.info('Saving private key to %s', key_file)
        with open(os.path.expanduser(key_file), 'w') as f1, open(
            os.path.expanduser(key_file) + '.pub', 'w'
        ) as f2:
            f1.write(private_key_str)
            f2.write(public_key_str)
            os.chmod(key_file, 0o600)

    return pub_key, priv_key


def get_or_generate_jwt_rsa_key_pair():
    """
    Get or generate a JWT RSA key pair.
    """
    from platypush.config import Config

    key_dir = os.path.join(Config.get_workdir(), 'jwt')
    priv_key_file = os.path.join(key_dir, 'id_rsa')
    pub_key_file = priv_key_file + '.pub'

    if os.path.isfile(priv_key_file) and os.path.isfile(pub_key_file):
        with open(pub_key_file, 'r') as f1, open(priv_key_file, 'r') as f2:
            return (
                PublicKey.load_pkcs1(f1.read().encode()),
                PrivateKey.load_pkcs1(f2.read().encode()),
            )

    pathlib.Path(key_dir).mkdir(parents=True, exist_ok=True, mode=0o755)
    return generate_rsa_key_pair(priv_key_file, size=2048)


def get_enabled_plugins() -> dict:
    """
    Get the enabled plugins.

    :return: A dictionary with the enabled plugins, in the format ``name`` ->
        :class:`platypush.plugins.Plugin` instance.
    """
    from platypush.config import Config
    from platypush.context import get_plugin

    plugins = {}
    for name in Config.get_plugins():
        try:
            plugin = get_plugin(name)
            if plugin:
                plugins[name] = plugin
        except Exception as e:
            logger.warning('Could not initialize plugin %s', name)
            logger.exception(e)

    return plugins


def get_enabled_backends() -> dict:
    """
    Get the enabled backends.

    :return: A dictionary with the enabled backends, in the format ``name`` ->
        :class:`platypush.backend.Backend` instance.
    """
    from platypush.config import Config
    from platypush.context import get_backend

    backends = {}
    for name in Config.get_backends():
        try:
            backend = get_backend(name.removeprefix('backend.'))
            if backend:
                backends[name] = backend
        except Exception as e:
            logger.warning('Could not initialize backend %s', name)
            logger.exception(e)

    return backends


def get_redis_pool(*args, **kwargs) -> ConnectionPool:
    """
    Get a Redis connection pool on the basis of the Redis configuration.

    The Redis configuration can be loaded from:

        1. The ``redis`` plugin.
        2. The ``backend.redis`` configuration (``redis_args`` attribute)

    """
    if not (args or kwargs):
        kwargs = get_redis_conf()

    pool_key = (kwargs.get('host', 'localhost'), kwargs.get('port', 6379))
    pool = redis_pools.get(pool_key)

    if not pool:
        pool = ConnectionPool(*args, **kwargs)
        redis_pools[pool_key] = pool

    return pool


def get_redis_conf() -> dict:
    """
    Get the Redis connection arguments from the configuration.
    """
    from platypush.config import Config

    return (
        Config.get('redis')
        or (Config.get('backend.redis') or {}).get('redis_args', {})
        or {}
    )


def get_redis(*args, **kwargs) -> Redis:
    """
    Get a Redis client on the basis of the Redis configuration.

    The Redis configuration can be loaded from:

        1. The ``redis`` plugin.
        2. The ``backend.redis`` configuration (``redis_args`` attribute)

    """
    return Redis(connection_pool=get_redis_pool(*args, **kwargs))


def to_datetime(t: Union[str, int, float, datetime.datetime]) -> datetime.datetime:
    """
    Utility function to convert a datetime/timestamp provided as a
    string/integer/float/datetime to a ``datetime.datetime`` instance.
    """
    if isinstance(t, (int, float)):
        return datetime.datetime.fromtimestamp(t, tz=tz.tzutc())
    if isinstance(t, str):
        return parser.parse(t)
    return t


@contextlib.contextmanager
def get_lock(
    lock: Lock, timeout: Optional[float] = None
) -> Generator[bool, None, None]:
    """
    Get a lock with an optional timeout through a context manager construct:

        >>> from threading import Lock
        >>> lock = Lock()
        >>> with get_lock(lock, timeout=2):
        >>>     ...

    """
    kwargs = {'timeout': timeout} if timeout else {}
    result = lock.acquire(**kwargs)

    try:
        if not result:
            raise TimeoutError()
        yield result
    finally:
        if result:
            lock.release()


def get_default_pid_file() -> str:
    """
    Get the default PID file path.
    """
    return os.path.join(gettempdir(), 'platypush.pid')


def get_remaining_timeout(
    timeout: Optional[float], start: float, cls: Union[Type[int], Type[float]] = float
) -> Optional[Union[int, float]]:
    """
    Get the remaining timeout, given a start time.
    """
    if timeout is None:
        return None

    return cls(max(0, timeout - (time.time() - start)))


def get_src_root() -> str:
    """
    :return: The root source folder of the application.
    """
    import platypush

    return os.path.dirname(inspect.getfile(platypush))


def is_root() -> bool:
    """
    :return: True if the current user is root/administrator.
    """
    return os.getuid() == 0


def get_message_response(msg):
    """
    Get the response to the given message.

    :param msg: The message to get the response for.
    :return: The response to the given message.
    """
    from platypush.message import Message

    redis = get_redis()
    redis_queue = get_redis_queue_name_by_message(msg)
    if not redis_queue:
        return None

    response = redis.blpop(redis_queue, timeout=60)
    if response and len(response) > 1:
        response = Message.build(response[1])
    else:
        response = None

    return response


def import_file(path: str, name: Optional[str] = None):
    """
    Import a Python file as a module, even if no __init__.py is
    defined in the directory.

    :param path: Path of the file to import.
    :param name: Custom name for the imported module (default: same as the file's basename).
    :return: The imported module.
    """
    name = name or re.split(r"\.py$", os.path.basename(path))[0]
    loader = SourceFileLoader(name, os.path.expanduser(path))
    mod_spec = spec_from_loader(name, loader)
    assert mod_spec, f"Cannot create module specification for {path}"
    mod = module_from_spec(mod_spec)
    loader.exec_module(mod)
    return mod


def get_defining_class(meth) -> Optional[type]:
    """
    See https://stackoverflow.com/a/25959545/622364.

    This is the best way I could find of answering the question "given a bound
    method, get the class that defined it",
    """
    if isinstance(meth, type):
        return meth

    if isinstance(meth, functools.partial):
        return get_defining_class(meth.func)

    if inspect.ismethod(meth) or (
        inspect.isbuiltin(meth)
        and getattr(meth, '__self__', None) is not None
        and getattr(meth.__self__, '__class__', None)
    ):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing

    if inspect.isfunction(meth):
        cls = getattr(
            inspect.getmodule(meth),
            meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
            None,
        )
        if isinstance(cls, type):
            return cls

    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def is_debug_enabled() -> bool:
    """
    :return: True if the debug mode is enabled.
    """
    from platypush.config import Config

    return (Config.get('logging') or {}).get('level') == logging.DEBUG


def get_default_downloads_dir() -> str:
    """
    :return: The default downloads directory.
    """
    return os.path.join(os.path.expanduser('~'), 'Downloads')


def wait_for_either(*events, timeout: Optional[float] = None, cls: Type = Event):
    """
    Wait for any of the given events to be set.

    :param events: The events to be checked.
    :param timeout: The maximum time to wait for the event to be set. Default: None.
    :param cls: The class to be used for the event. Default: threading.Event.
    """
    from .threads import OrEvent

    return OrEvent(*events, cls=cls).wait(timeout=timeout)


def utcnow():
    """
    utcnow() without tears. It always returns a datetime object in UTC
    timezone.
    """
    return datetime.datetime.now(datetime.timezone.utc)


# vim:sw=4:ts=4:et:



from logging import getLogger
from queue import Empty, Queue
from typing import Callable, Optional

import sounddevice as sd


# pylint: disable=too-few-public-methods
class AudioOutputCallback:
    """
    The ``AudioSynthOutput`` is a functor that wraps the ``sounddevice.Stream``
    callback and writes raw audio data to the audio device.
    """

    def __init__(
        self,
        *args,
        audio_queue: Queue,  # Queue[NDArray[np.number]],
        channels: int,
        blocksize: int,
        should_stop: Callable[[], bool] = lambda: False,
        is_paused: Callable[[], bool] = lambda: False,
        queue_timeout: Optional[float] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._audio_queue = audio_queue
        self._channels = channels
        self._blocksize = blocksize
        self._should_stop = should_stop
        self._is_paused = is_paused
        self._queue_timeout = queue_timeout
        self.logger = getLogger(__name__)

    def _check_status(self, frames: int, status):
        """
        Checks the current status of the audio callback and raises errors if
        the processing shouldn't continue.
        """
        if self._should_stop():
            raise sd.CallbackStop

        assert frames == self._blocksize, (
            f'Received {frames} frames, expected blocksize is {self._blocksize}',
        )

        assert not status.output_underflow, 'Output underflow: increase blocksize?'
        assert not status, f'Audio callback failed: {status}'

    # outdata: NDArray[np.number]
    def _audio_callback(self, outdata, frames: int, status):
        if self._is_paused():
            return

        self._check_status(frames, status)

        try:
            data = self._audio_queue.get_nowait()
        except Empty as e:
            raise (
                sd.CallbackStop
                if self._should_stop()
                else AssertionError('Buffer is empty: increase buffersize?')
            ) from e

        if data.shape[0] == 0:
            raise sd.CallbackStop

        audio_length = min(len(data), len(outdata))
        outdata[:audio_length] = data[:audio_length]

    # _ = time
    # outdata: NDArray[np.number]
    def __call__(self, outdata, frames: int, _, status):
        try:
            self._audio_callback(outdata, frames, status)
        except AssertionError as e:
            self.logger.warning(str(e))



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class TodoistUserSchema(Schema):
    """
    Todoist user schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    auto_reminder = fields.Int()
    avatar_big = fields.Url(
        metadata={
            "description": "The user's avatar URL.",
            "example": "https://example.com/user/100x100.png",
        },
    )

    avatar_medium = fields.Url(
        metadata={
            "description": "The user's avatar URL.",
            "example": "https://example.com/user/50x50.png",
        },
    )

    avatar_s640 = fields.Url(
        metadata={
            "description": "The user's avatar URL.",
            "example": "https://example.com/user/640x640.png",
        },
    )

    avatar_small = fields.Url(
        metadata={
            "description": "The user's avatar URL.",
            "example": "https://example.com/user/25x25.png",
        },
    )

    business_account_id = fields.Int(
        metadata={
            "description": "The user's business account ID.",
            "example": 123456,
        },
    )

    daily_goal = fields.Int(
        metadata={
            "description": "The user's daily goal.",
            "example": 100,
        },
    )

    date_format = fields.String(
        metadata={
            "description": "The user's date format.",
            "example": "dd-mm-yyyy",
        },
    )

    dateist_inline_disabled = fields.Bool()
    dateist_lang = fields.String(
        metadata={
            "description": "The user's dateist language.",
            "example": "en",
        },
    )

    days_off = fields.List(
        fields.Int(),
        metadata={
            "description": "The user's days off.",
            "example": [0, 6],
        },
    )

    default_reminder = fields.String()
    email = fields.Email(
        metadata={
            "description": "The user's email.",
            "example": "user@example.com",
        },
    )

    features = fields.Dict()
    full_name = fields.String(
        metadata={
            "description": "The user's full name.",
            "example": "John Doe",
        },
    )

    id = fields.Int(
        required=True,
        metadata={
            "description": "The user's unique identifier.",
            "example": 123456,
        },
    )

    image_id = fields.String()
    inbox_project = fields.Int(
        metadata={
            "description": "The user's inbox project ID.",
            "example": 123456,
        },
    )

    is_biz_admin = fields.Bool()
    is_premium = fields.Bool()
    join_date = DateTime()
    karma = fields.Float()
    karma_trend = fields.String()
    lang = fields.String(
        metadata={
            "description": "The user's language.",
            "example": "en",
        },
    )

    legacy_inbox_project = fields.Int()
    mobile_host = fields.String()
    mobile_number = fields.String()
    next_week = fields.Int()
    premium_until = DateTime()
    share_limit = fields.Int()
    sort_order = fields.Int()
    start_day = fields.Int()
    start_page = fields.String()
    theme = fields.Int()
    time_format = fields.String(
        metadata={
            "description": "The user's time format.",
            "example": "24h",
        },
    )

    token = fields.String()
    tz_info = fields.Dict()
    unique_prefix = fields.Int()
    websocket_url = fields.Url()
    weekly_goal = fields.Int()


class TodoistProjectSchema(Schema):
    """
    Todoist project schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    child_order = fields.Int()
    collapsed = fields.Bool()
    color = fields.Int()
    has_more_notes = fields.Bool()
    id = fields.Int(required=True)
    is_archived = fields.Bool()
    is_deleted = fields.Bool()
    is_favorite = fields.Bool()
    name = fields.String(required=True)
    shared = fields.Bool()
    inbox_project = fields.Bool()
    legacy_id = fields.Int()
    parent_id = fields.Int()


class TodoistItemSchema(Schema):
    """
    Todoist item schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    assigned_by_uid = fields.Int()
    checked = fields.Bool()
    child_order = fields.Int()
    collapsed = fields.Bool()
    content = fields.String()
    date_added = DateTime()
    date_completed = DateTime()
    day_order = fields.Int()
    due = fields.Dict()
    has_more_notes = fields.Bool()
    id = fields.Int(required=True)
    in_history = fields.Bool()
    is_deleted = fields.Bool()
    labels = fields.List(fields.String())
    legacy_project_id = fields.Int()
    parent_id = fields.Int()
    priority = fields.Int()
    project_id = fields.Int()
    responsible_uid = fields.Int()
    section_id = fields.Int()
    sync_id = fields.Int()
    user_id = fields.Int()


class TodoistFilterSchema(Schema):
    """
    Todoist filter schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    color = fields.Int()
    id = fields.Int(required=True)
    is_deleted = fields.Bool()
    item_order = fields.Int()
    name = fields.String()
    query = fields.String()
    user_id = fields.Int()


class TodoistLiveNotificationSchema(Schema):
    """
    Todoist live notifications schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    id = fields.Int(required=True)
    notification_key = fields.String()
    notification_type = fields.String()
    project_id = fields.Int()
    user_id = fields.Int()
    is_deleted = fields.Bool()
    is_unread = fields.Bool()
    completed_last_month = fields.Bool()
    karma_level = fields.Int()
    promo_img = fields.Url()
    completed_tasks = fields.Int()
    created = fields.String()


class TodoistCollaboratorSchema(Schema):
    """
    Todoist collaborator schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    data = fields.Dict()


class TodoistNoteSchema(Schema):
    """
    Todoist note schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    data = fields.Dict()


class TodoistProjectNoteSchema(Schema):
    """
    Todoist project note schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    data = fields.Dict()



class CameraException(RuntimeError):
    pass


class CaptureAlreadyRunningException(CameraException):
    def __init__(self, device):
        super().__init__('A capturing session on the device {} is already running'.format(device))


# vim:sw=4:ts=4:et:



from abc import ABC
from threading import Event, Thread
from typing import Optional


class StoppableThread(Thread, ABC):
    """
    Base interface for stoppable threads.
    """

    def __init__(self, *args, stop_event: Optional[Event] = None, **kwargs):
        """
        :param stop_event: Event used to signal the thread to stop.
        """
        super().__init__(*args, **kwargs)
        self._stop_event = stop_event or Event()

    def should_stop(self) -> bool:
        """
        :return: ``True`` if the thread should be stopped, ``False`` otherwise.
        """
        return self._stop_event.is_set()

    def wait_stop(self, timeout: Optional[float] = None):
        """
        Wait for the stop event to be set.

        :param timeout: Timeout in seconds (default: no timeout).
        """
        self._stop_event.wait(timeout)

    def stop(self):
        """
        Signal the thread to stop by setting the stop event.
        """
        self._stop_event.set()



from .._base import BaseBluetoothClass, ClassProperty


class MajorDeviceClass(BaseBluetoothClass):
    """
    Models Bluetooth major device classes - see
    https://btprodspecificationrefs.blob.core.windows.net/assigned-numbers/Assigned%20Number%20Types/Assigned%20Numbers.pdf,
    Section 2.8.2
    """

    UNKNOWN = ClassProperty('Unknown', 0x1F00, 8, 0b0000)
    COMPUTER = ClassProperty('Computer', 0x1F00, 8, 0b0001)
    PHONE = ClassProperty('Phone', 0x1F00, 8, 0b0010)
    AP = ClassProperty('LAN / Network Access Point', 0x1F00, 8, 0b0011)
    MULTIMEDIA = ClassProperty('Audio / Video', 0x1F00, 8, 0b0100)
    PERIPHERAL = ClassProperty('Peripheral', 0x1F00, 8, 0b0101)
    IMAGING = ClassProperty('Imaging', 0x1F00, 8, 0b0110)
    WEARABLE = ClassProperty('Wearable', 0x1F00, 8, 0b0111)
    TOY = ClassProperty('Toy', 0x1F00, 8, 0b1000)
    HEALTH = ClassProperty('Health', 0x1F00, 8, 0b1001)



from sqlalchemy import Column, Integer, String, ForeignKey, Boolean, Float

from platypush.common.db import is_defined

from .devices import Device


if not is_defined('light'):

    class Light(Device):
        __tablename__ = 'light'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        on = Column(Boolean)
        brightness = Column(Float)
        saturation = Column(Float)
        hue = Column(Float)
        temperature = Column(Float)
        x = Column(Float)
        y = Column(Float)
        red = Column(Float)
        green = Column(Float)
        blue = Column(Float)
        colormode = Column(String)
        effect = Column(String)
        hue_min = Column(Float)
        hue_max = Column(Float)
        saturation_min = Column(Float)
        saturation_max = Column(Float)
        brightness_min = Column(Float)
        brightness_max = Column(Float)
        temperature_min = Column(Float)
        temperature_max = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from queue import Queue, Empty
from threading import Thread
from time import time
from traceback import format_exception
from typing import Optional, Any, Collection, Mapping

from sqlalchemy import or_, text
from sqlalchemy.orm import make_transient, Session

from platypush.config import Config
from platypush.context import get_plugin, get_bus
from platypush.entities import (
    Entity,
    get_plugin_entity_registry,
    get_entities_registry,
)
from platypush.message.event.entities import EntityUpdateEvent, EntityDeleteEvent
from platypush.plugins import Plugin, action


class EntitiesPlugin(Plugin):
    """
    This plugin is used to interact with native platform entities (e.g. switches, lights,
    sensors etc.) through a consistent interface, regardless of the integration type.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_session(self, *args, **kwargs) -> Session:
        db = get_plugin('db')
        assert db
        return db.get_session(*args, **kwargs)

    @action
    def get(
        self,
        types: Optional[Collection[str]] = None,
        plugins: Optional[Collection[str]] = None,
        **filter,
    ):
        """
        Retrieve a list of entities.

        :param types: Entity types, as specified by the (lowercase) class name and table name.
            Default: all entities.
        :param plugins: Filter by plugin IDs (default: all plugins).
        :param filter: Filter entities with these criteria (e.g. `name`, `id`,
            `state`, `type`, `plugin` etc.)
        """
        entity_registry = get_entities_registry()
        selected_types = []
        all_types = {
            e.__tablename__.lower(): e for e in entity_registry  # type: ignore
        }

        if types:
            selected_types = {t.lower() for t in types}
            entity_types = {t: et for t, et in all_types.items() if t in selected_types}
            invalid_types = selected_types.difference(entity_types.keys())
            assert not invalid_types, (
                f'No such entity types: {invalid_types}. '
                f'Supported types: {list(all_types.keys())}'
            )

            selected_types = entity_types.keys()

        enabled_plugins = list(
            {
                *Config.get_plugins().keys(),
                *Config.get_backends().keys(),
            }
        )

        with self._get_session() as session:
            query = session.query(Entity).filter(
                or_(Entity.plugin.in_(enabled_plugins), Entity.plugin.is_(None))
            )

            if selected_types:
                query = query.filter(Entity.type.in_(selected_types))
            if plugins:
                query = query.filter(Entity.plugin.in_(plugins))
            if filter:
                query = query.filter_by(**filter)

            return [e.to_json() for e in query.all()]

    @action
    def scan(
        self,
        types: Optional[Collection[str]] = None,
        plugins: Optional[Collection[str]] = None,
        timeout: Optional[float] = 30.0,
    ):
        """
        (Re-)scan entities and return the updated results.

        :param types: Filter by entity types (e.g. `switch`, `light`, `sensor` etc.).
        :param plugins: Filter by plugin names (e.g. `switch.tplink` or `light.hue`).
        :param timeout: Scan timeout in seconds. Default: 30.
        """
        filter = {}
        plugin_registry = get_plugin_entity_registry()

        if plugins:
            filter['plugins'] = plugins
            plugin_registry['by_plugin'] = {
                plugin: plugin_registry['by_plugin'][plugin]
                for plugin in plugins
                if plugin in plugin_registry['by_plugin']
            }

        if types:
            filter['types'] = types
            filter_entity_types = set(types)
            plugin_registry['by_plugin'] = {
                plugin_name: entity_types
                for plugin_name, entity_types in plugin_registry['by_plugin'].items()
                if any(t for t in entity_types if t in filter_entity_types)
            }

        enabled_plugins = plugin_registry['by_plugin'].keys()

        def worker(plugin_name: str, q: Queue):
            try:
                plugin = get_plugin(plugin_name)
                assert plugin, f'No such configured plugin: {plugin_name}'
                # Force a plugin scan by calling the `status` action
                response = plugin.status()
                assert not response.errors, response.errors
                q.put((plugin_name, response.output))
            except Exception as e:
                q.put((plugin_name, e))

        q = Queue()
        start_time = time()
        results = []
        workers = [
            Thread(target=worker, args=(plugin_name, q))
            for plugin_name in enabled_plugins
        ]

        for w in workers:
            w.start()

        while len(results) < len(workers) and (
            not timeout or (time() - start_time < timeout)
        ):
            try:
                plugin_name, result = q.get(block=True, timeout=0.5)
                if isinstance(result, Exception):
                    self.logger.warning(
                        'Could not load results from plugin %s: %s', plugin_name, result
                    )

                    self.logger.warning(
                        ''.join(
                            format_exception(
                                type(result), value=result, tb=result.__traceback__
                            )
                        )
                    )
                else:
                    results.append(result)
            except Empty:
                continue

        if len(results) < len(workers):
            self.logger.warning('Scan timed out for some plugins')

        for w in workers:
            w.join(timeout=max(0, timeout - (time() - start_time)) if timeout else None)

        return self.get(**filter)

    @action
    def execute(self, id: Any, action: str, *args, **kwargs):
        """
        Execute an action on an entity (for example `on`/`off` on a switch, or `get`
        on a sensor).

        :param id: Entity ID (i.e. the entity's db primary key, not the plugin's external
            or "logical" key)
        :param action: Action that should be run. It should be a method implemented
            by the entity's class.
        :param args: Action's extra positional arguments.
        :param kwargs: Action's extra named arguments.
        """
        with self._get_session() as session:
            entity = session.query(Entity).filter_by(id=id).one_or_none()

        assert entity, f'No such entity ID: {id}'
        return entity.run(action, *args, **kwargs)

    @action
    def delete(self, *entities: int):  # type: ignore
        """
        Delete a set of entity IDs.

        Note: this should only be done if the entity is no longer available or
        the associated plugin has been disabled, otherwise the entities will be
        re-created by the plugins on the next scan.

        :param entities: IDs of the entities to be removed.
        :return: The payload of the deleted entities.
        """
        with self._get_session(locked=True) as session:
            if str(session.connection().engine.url).startswith('sqlite://'):
                # SQLite requires foreign_keys to be explicitly enabled
                # in order to proper manage cascade deletions
                session.execute(text('PRAGMA foreign_keys = ON'))

            entities: Collection[Entity] = (
                session.query(Entity).filter(Entity.id.in_(entities)).all()
            )
            for entity in entities:
                session.delete(entity)
            session.commit()

        for entity in entities:
            make_transient(entity)
            get_bus().post(EntityDeleteEvent(entity))

        return entities

    @action
    def rename(self, **entities: Mapping[str, str]):
        """
        Rename a sequence of entities.
        Renaming, as of now, is actually done by setting the ``.meta.name_override``
        property of an entity rather than fully renaming the entity (which may be owned
        by a plugin that doesn't support renaming, therefore the next entity update may
        overwrite the name).

        :param entities: Entity `id` -> `new_name` mapping.
        """
        return self.set_meta(
            **{
                entity_id: {'name_override': name}
                for entity_id, name in entities.items()
            }
        )

    @action
    def set_meta(self, **entities):
        """
        Update the metadata of a set of entities.

        :param entities: Entity `id` -> `new_metadata_fields` mapping.
        :return: The updated entities.
        """
        entities = {str(k): v for k, v in entities.items()}
        with self._get_session(locked=True) as session:
            objs = session.query(Entity).filter(Entity.id.in_(entities.keys())).all()
            for obj in objs:
                obj.meta = {  # type: ignore
                    **dict(obj.meta or {}),  # type: ignore
                    **(entities.get(str(obj.id), {})),
                }
                session.add(obj)

            session.commit()

        for obj in objs:
            make_transient(obj)
            get_bus().post(EntityUpdateEvent(obj))

        return [obj.to_json() for obj in objs]


# vim:sw=4:ts=4:et:



import enum
import time

from typing import List, Optional, Dict, Union, Callable, Tuple, Type

from pyfirmata2 import (
    Arduino,
    ArduinoMega,
    ArduinoDue,
    ArduinoNano,
    Board,
    Pin,
    util,
    ANALOG,
    INPUT,
    PWM,
)

from platypush.common.sensors import Numeric
from platypush.entities.devices import Device
from platypush.entities.sensors import NumericSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


class PinType(enum.IntEnum):
    """
    PIN type enumeration (analog or digital).
    """

    ANALOG = 1
    DIGITAL = 2


class BoardType(enum.Enum):
    """
    Board types.
    """

    MEGA = 'mega'
    DUE = 'due'
    NANO = 'nano'


# pylint: disable=too-many-ancestors
class ArduinoPlugin(SensorPlugin):
    """
    Interact with an Arduino connected to the host machine over USB using the
    `Firmata <https://www.arduino.cc/en/reference/firmata>`_ protocol.

    You have two options to communicate with an Arduino-compatible board over USB:

        - Use this plugin if you want to use the general-purpose Firmata protocol - in this case most of your
          processing logic will be on the host side and you can read/write data to the Arduino transparently.
        - Use the :class:`platypush.plugins.serial.SerialPlugin` if instead you want to run more custom logic
          on the Arduino and communicate back with the host computer through JSON formatted messages.

    Download and flash the
    `Standard Firmata <https://github.com/firmata/arduino/blob/master/examples/StandardFirmata/StandardFirmata.ino>`_
    firmware to the Arduino in order to use this plugin.
    """

    def __init__(
        self,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: int = 9600,
        analog_pins: Optional[Dict[str, int]] = None,
        digital_pins: Optional[Dict[str, int]] = None,
        timeout: float = 20.0,
        conv_functions: Optional[Dict[Union[str, int], Union[str, Callable]]] = None,
        poll_interval: float = 1.0,
        **kwargs,
    ):
        """
        :param board: Default board name or path (e.g. ``COM3`` on Windows or ``/dev/ttyUSB0`` on Unix). If not set
            then the plugin will attempt an auto-discovery.

        :param board_type: Default board type. It can be 'mega', 'due' or 'nano'. Leave empty for auto-detection.
        :param baud_rate: Default serial baud rate (default: 9600)
        :param analog_pins: Optional analog PINs map name->pin_number.
        :param digital_pins: Optional digital PINs map name->pin_number.
        :param timeout: Board communication timeout in seconds.
        :param conv_functions: Optional mapping of conversion functions to apply to the analog values read from a
            certain PIN. The key can either be the PIN number or the name as specified in ``analog_pins``, the value
            can be a function that takes an argument and transforms it or its lambda string representation.
            Note that ``analog_read`` returns by default float values in the range [0.0, 1.0]. Example:

            .. code-block:: yaml

                arduino:
                    board: /dev/ttyUSB0
                    analog_pins:
                        temperature: 1  # Analog PIN 1

                    conv_functions:
                        temperature: 'lambda t: t * 500.0'

        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        self.board = board
        self.board_type = self._get_board_type(board_type)
        self.baud_rate = baud_rate
        self.timeout = timeout

        self._pin_number_by_name = {
            PinType.ANALOG: analog_pins or {},
            PinType.DIGITAL: digital_pins or {},
        }

        self._pin_name_by_number = {
            PinType.ANALOG: {
                number: name
                for name, number in self._pin_number_by_name[PinType.ANALOG].items()
            },
            PinType.DIGITAL: {
                number: name
                for name, number in self._pin_number_by_name[PinType.DIGITAL].items()
            },
        }

        self.conv_functions: Dict[Union[str, int], Callable] = {
            (self._pin_number_by_name[PinType.ANALOG].get(str(pin), pin)): (
                f if callable(f) else eval(f)
            )
            for pin, f in (conv_functions or {}).items()
        }

        self._boards: Dict[str, Board] = {}
        self._board_iterators: Dict[str, util.Iterator] = {}

    @staticmethod
    def _get_board_type(board_type: Optional[str] = None) -> Type[Board]:
        if not board_type:
            return Arduino

        board_type = board_type.lower()
        if board_type == BoardType.DUE.value:
            return ArduinoDue
        if board_type == BoardType.NANO.value:
            return ArduinoNano
        if board_type == BoardType.MEGA.value:
            return ArduinoMega

        raise AssertionError(f'Invalid board_type: {board_type}')

    def _get_board(
        self,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[float] = None,
    ) -> Board:
        board_name = board or self.board or Arduino.AUTODETECT
        baud_rate = baud_rate or self.baud_rate
        timeout = timeout or self.timeout

        if board_name in self._boards:
            return self._boards[board_name]

        board_obj_type = (
            self._get_board_type(board_type) if board_type else self.board_type
        )
        assert board_obj_type

        board_obj = board_obj_type(board_name, baudrate=baud_rate, timeout=timeout)
        board_name = board_obj.name or ''
        self.logger.info('Connected to board %s', board_name)
        self._boards[board_name] = board_obj
        self._board_iterators[board_name] = util.Iterator(board_obj)
        self._board_iterators[board_name].start()
        return board_obj

    def _get_board_and_pin(
        self,
        pin: Union[int, str],
        pin_type: PinType,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[int] = None,
    ) -> Tuple[Board, int]:
        board_ = self._get_board(
            board, board_type=board_type, baud_rate=baud_rate, timeout=timeout
        )
        if pin in self._pin_number_by_name[pin_type]:
            pin = self._pin_number_by_name[pin_type][str(pin)]

        assert isinstance(pin, int), f'Invalid PIN number/name: {pin}'
        return board_, pin

    @staticmethod
    def _get_pin(pin: int, board: Board, pin_type: PinType) -> Pin:
        pins = None
        if pin_type == PinType.ANALOG:
            pins = board.analog
        if pin_type == PinType.DIGITAL:
            pins = board.digital

        assert pins, f'Invalid pin_type: {pin_type}'

        if pins[pin].mode in [ANALOG, INPUT]:
            pins[pin].enable_reporting()
        return pins[pin]

    def _poll_value(
        self,
        pin: int,
        board: Board,
        pin_type: PinType,
        timeout: Optional[float] = None,
    ) -> Optional[Union[bool, float]]:
        value = None
        poll_start = time.time()

        while value is None:
            if timeout and time.time() - poll_start >= timeout:
                raise RuntimeError('Read timeout')

            pin_ = self._get_pin(pin=pin, board=board, pin_type=pin_type)
            if pin_.mode not in [INPUT, ANALOG]:
                self.logger.warning(
                    'PIN %d is not configured in input/analog mode', pin
                )
                return None

            value = pin_.read()
            if value is None:
                time.sleep(0.001)

        if pin_type == PinType.DIGITAL:
            value = bool(value)

        return value

    @action
    def analog_read(
        self,
        pin: Union[int, str],
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        conv_function: Optional[Union[str, Callable]] = None,
        timeout: Optional[int] = None,
    ) -> Optional[float]:
        """
        Read an analog value from a PIN.

        :param pin: PIN number or configured name.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param conv_function: Optional conversion function override to apply to the output. It can be either a function
            object or its lambda string representation (e.g. ``lambda x: x*x``). Keep in mind that ``analog_read``
            returns by default float values in the range ``[0.0, 1.0]``.
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board_, pin = self._get_board_and_pin(
            pin=pin,
            pin_type=PinType.ANALOG,
            board=board,
            board_type=board_type,
            baud_rate=baud_rate,
            timeout=timeout,
        )

        conv_function = conv_function or self.conv_functions.get(pin)
        converter: Optional[Callable[[float], float]] = None
        if isinstance(conv_function, str):
            converter = eval(conv_function)
        elif callable(conv_function):
            converter = conv_function
        elif conv_function is not None:
            raise AssertionError(
                'Expected conv_function to be null, a string or a function, '
                f'got "{conv_function}" instead'
            )

        value = self._poll_value(
            pin=pin, board=board_, pin_type=PinType.ANALOG, timeout=timeout
        )

        if converter and value is not None:
            value = converter(value)
        return value

    @action
    def digital_read(
        self,
        pin: Union[int, str],
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Read a digital value from a PIN.

        :param pin: PIN number or configured name.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board_, pin = self._get_board_and_pin(
            pin=pin,
            pin_type=PinType.DIGITAL,
            board=board,
            board_type=board_type,
            baud_rate=baud_rate,
            timeout=timeout,
        )

        return bool(
            self._poll_value(
                pin=pin, board=board_, pin_type=PinType.DIGITAL, timeout=timeout
            )
        )

    @action
    def analog_write(
        self,
        pin: Union[int, str],
        value: float,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[int] = None,
    ):
        """
        Write a value to an analog PIN.

        :param pin: PIN number or configured name.
        :param value: Voltage to be sent, a real number normalized between 0 and 1.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board_, pin = self._get_board_and_pin(
            pin=pin,
            pin_type=PinType.ANALOG,
            board=board,
            board_type=board_type,
            baud_rate=baud_rate,
            timeout=timeout,
        )
        board_.analog[pin].write(value)

    @action
    def digital_write(
        self,
        pin: Union[int, str],
        value: bool,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[int] = None,
    ):
        """
        Write a value to a digital PIN.

        :param pin: PIN number or configured name.
        :param value: True (HIGH) or False (LOW).
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board_, pin = self._get_board_and_pin(
            pin=pin,
            pin_type=PinType.DIGITAL,
            board=board,
            board_type=board_type,
            baud_rate=baud_rate,
            timeout=timeout,
        )
        board_.digital[pin].write(value)

    @action
    def pwm_write(
        self,
        pin: Union[int, str],
        value: float,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[int] = None,
    ):
        """
        Write a PWM value to a digital PIN.

        :param pin: PIN number or configured name.
        :param value: PWM real value normalized between 0 and 1.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board_, pin = self._get_board_and_pin(
            pin=pin,
            pin_type=PinType.DIGITAL,
            board=board,
            board_type=board_type,
            baud_rate=baud_rate,
            timeout=timeout,
        )

        assert board_.digital[pin].PWM_CAPABLE, f'PIN {pin} is not PWM capable'
        if board_.digital[pin] != PWM:
            board_.digital[pin].mode = PWM
            time.sleep(0.001)  # 1 μs spike to activate a PWM pin

        board_.digital[pin].write(value)

    @action
    def get_measurement(
        self,
        *_,
        board: Optional[str] = None,
        board_type: Optional[str] = None,
        baud_rate: Optional[int] = None,
        timeout: Optional[int] = None,
        **__,
    ) -> Dict[str, Optional[Union[float, bool]]]:
        """
        Get a measurement from all the configured PINs.

        :param board: Board path or name (default: default configured ``board``)
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``)
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        :return: dict, where the keys are either the configured names of the PINs (see ``analog_pins`` configuration)
            or all the analog PINs (names will be in the format 'A0..A7' in that case), and the values will be the
            real values measured, either normalized between 0 and 1 if no conversion functions were provided, or
            transformed through the configured ``conv_functions``.
        """
        ret = {}
        board_ = self._get_board(
            board=board, board_type=board_type, baud_rate=baud_rate, timeout=timeout
        )

        assert board_, f'No such board: board={board}, board_type={board_type}'

        for pin in board_.analog:
            if (
                self._pin_name_by_number[PinType.ANALOG]
                and pin.pin_number not in self._pin_name_by_number[PinType.ANALOG]
            ):
                continue

            name = self._pin_name_by_number[PinType.ANALOG].get(
                pin.pin_number, f'A{pin.pin_number}'
            )
            value = self._poll_value(
                pin=pin.pin_number,
                board=board_,
                pin_type=PinType.ANALOG,
                timeout=timeout or self.timeout,
            )

            if value is None:
                continue

            conv_function = self.conv_functions.get(
                name, self.conv_functions.get(pin.pin_number)
            )
            if conv_function:
                value = conv_function(value)

            ret[name] = value

        return ret

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:  # type: ignore
        dev_id = 'arduino'
        dev_name = 'Arduino'
        if self.board:
            dev_id += f':{self.board}'
            dev_name += f' @ {self.board}'

        return [
            Device(
                id=dev_id,
                name=dev_name,
                children=[
                    NumericSensor(
                        id=f'{dev_id}:{key}',
                        name=key,
                        value=value,
                    )
                    for key, value in entities.items()
                ],
            )
        ]

    @action
    def stop(self):
        super().stop()
        for it in self._board_iterators.values():
            it.stop()

        for board in self._boards.values():
            board.exit()

        self._board_iterators = {}
        self._boards = {}


# vim:sw=4:ts=4:et:



import base64
import json

from platypush.message.event.nfc import (
    NFCTagDetectedEvent,
    NFCTagRemovedEvent,
    NFCDeviceConnectedEvent,
    NFCDeviceDisconnectedEvent,
)
from platypush.plugins import RunnablePlugin, action


class NfcPlugin(RunnablePlugin):
    """
    Plugin to detect events from NFC devices.

    Run the following command to check if your device is compatible with nfcpy
    and the right permissions are set::

        python -m nfc

    """

    def __init__(self, device='usb', **kwargs):
        """
        :param device: Address or ID of the device to be opened. Examples:

            * ``usb:003:009`` opens device 9 on bus 3
            * ``usb:003`` opens the first available device on bus 3
            * ``usb`` opens the first available USB device (default)
        """
        super().__init__(**kwargs)
        self.device_id = device
        self._clf = None

    def _get_clf(self):
        import nfc

        if not self._clf:
            self._clf = nfc.ContactlessFrontend()
            self._clf.open(self.device_id)
            assert self._clf and self._clf.device, 'NFC reader not initialized'
            self._bus.post(NFCDeviceConnectedEvent(reader=self._get_device_str()))
            self.logger.info(
                'Initialized NFC reader on device %s', self._get_device_str()
            )

        return self._clf

    def _get_device_str(self):
        assert self._clf, 'NFC reader not initialized'
        return str(self._clf.device)

    def close(self):
        if self._clf:
            self._clf.close()
            self._clf = None
            self._bus.post(NFCDeviceDisconnectedEvent(reader=self._get_device_str()))

    @staticmethod
    def _parse_records(tag):
        from ndef.text import TextRecord
        from ndef.uri import UriRecord
        from ndef.smartposter import SmartposterRecord
        from ndef.deviceinfo import DeviceInformationRecord
        from ndef.wifi import WifiSimpleConfigRecord, WifiPeerToPeerRecord
        from ndef.bluetooth import BluetoothLowEnergyRecord, BluetoothEasyPairingRecord
        from ndef.signature import SignatureRecord

        records = []

        if not tag.ndef:
            return records

        for record in tag.ndef.records:
            r = {
                'record_type': record.type,
                'record_name': record.name,
            }

            if isinstance(record, TextRecord):
                try:
                    r = {
                        **r,
                        'type': 'json',
                        'value': json.loads(record.text),
                    }
                except ValueError:
                    r = {
                        **r,
                        'type': 'text',
                        'text': record.text,
                    }
            elif isinstance(record, UriRecord):
                r = {
                    **r,
                    'type': 'uri',
                    'uri': record.uri,
                    'iri': record.iri,
                }
            elif isinstance(record, SmartposterRecord):
                r = {
                    **r,
                    'type': 'smartposter',
                    **{
                        attr: getattr(record, attr)
                        for attr in [
                            'resource',
                            'titles',
                            'title',
                            'action',
                            'icon',
                            'icons',
                            'resource_size',
                            'resource_type',
                        ]
                    },
                }
            elif isinstance(record, DeviceInformationRecord):
                r = {
                    **r,
                    'type': 'device_info',
                    **{
                        attr: getattr(record, attr)
                        for attr in [
                            'vendor_name',
                            'model_name',
                            'unique_name',
                            'uuid_string',
                            'version_string',
                        ]
                    },
                }
            elif isinstance(record, WifiSimpleConfigRecord):
                r = {
                    **r,
                    'type': 'wifi_simple_config',
                    **{attr: record[attr] for attr in record.attribute_names()},
                }
            elif isinstance(record, WifiPeerToPeerRecord):
                r = {
                    **r,
                    'type': 'wifi_peer_to_peer',
                    **{attr: record[attr] for attr in record.attribute_names()},
                }
            elif isinstance(record, BluetoothEasyPairingRecord):
                r = {
                    **r,
                    'type': 'bluetooth_easy_pairing',
                    **{
                        attr: getattr(record, attr)
                        for attr in ['device_address', 'device_name', 'device_class']
                    },
                }
            elif isinstance(record, BluetoothLowEnergyRecord):
                r = {
                    **r,
                    'type': 'bluetooth_low_energy',
                    **{
                        attr: getattr(record, attr)
                        for attr in [
                            'device_address',
                            'device_name',
                            'role_capabilities',
                            'appearance',
                            'flags',
                            'security_manager_tk_value',
                            'secure_connections_confirmation_value',
                            'secure_connections_random_value',
                        ]
                    },
                }
            elif isinstance(record, SignatureRecord):
                r = {
                    **r,
                    'type': 'signature',
                    **{
                        attr: getattr(record, attr)
                        for attr in [
                            'version',
                            'signature_type',
                            'hash_type',
                            'signature',
                            'signature_uri',
                            'certificate_format',
                            'certificate_store',
                            'certificate_uri',
                            'secure_connections_random_value',
                        ]
                    },
                }
            else:
                r = {
                    **r,
                    'type': 'binary',
                    'data': base64.encodebytes(record.data).decode(),
                }

            records.append(r)

        return records

    @staticmethod
    def _parse_id(tag):
        return ''.join([('%02X' % c) for c in tag.identifier])

    def _on_connect(self):
        def callback(tag):
            if not tag:
                return False

            tag_id = self._parse_id(tag)
            records = self._parse_records(tag)
            self._bus.post(
                NFCTagDetectedEvent(
                    reader=self._get_device_str(), tag_id=tag_id, records=records
                )
            )
            return True

        return callback

    def _on_release(self):
        def callback(tag):
            tag_id = self._parse_id(tag)
            self._bus.post(
                NFCTagRemovedEvent(reader=self._get_device_str(), tag_id=tag_id)
            )

        return callback

    @action
    def status(self):
        """
        Get information about the NFC reader status.

        Example output:

          .. code-block:: json

            {
                "display_name": "ACS ACR122U PN532v1.6 at usb:001:017",
                "path": "usb:001:017",
                "product_name": "ACR122U",
                "vendor_name": "ACS"
            }

        """
        assert self._clf and self._clf.device, 'NFC reader not initialized'
        return {
            'display_name': str(self._clf.device),
            'path': self._clf.device.path,
            'product_name': self._clf.device.product_name,
            'vendor_name': self._clf.device.vendor_name,
        }

    def main(self):
        fail_wait = 5
        max_fail_wait = 60

        while not self.should_stop():
            try:
                clf = self._get_clf()
                clf.connect(
                    rdwr={
                        'on-connect': self._on_connect(),
                        'on-release': self._on_release(),
                    }
                )
                fail_wait = 5
            except Exception as e:
                self.logger.warning(
                    'NFC error, retrying in %d seconds: %s', e, fail_wait, exc_info=True
                )
                self.wait_stop(fail_wait)
                fail_wait = min(fail_wait * 2, max_fail_wait)
            finally:
                self.close()


# vim:sw=4:ts=4:et:



import re
import threading
from typing import Collection, Optional, Union

from platypush.plugins import RunnablePlugin, action
from platypush.plugins.music import MusicPlugin

from ._conf import MpdConfig
from ._listener import MpdListener


class MusicMpdPlugin(MusicPlugin, RunnablePlugin):
    """
    This plugin allows you to interact with an MPD/Mopidy music server.

    `MPD <https://www.musicpd.org/>`_ is a flexible server-side
    protocol/application for handling music collections and playing music,
    mostly aimed to manage local libraries.

    `Mopidy <https://www.mopidy.com/>`_ is an evolution of MPD, compatible with
    the original protocol and with support for multiple music sources through
    plugins (e.g. Spotify, TuneIn, Soundcloud, local files etc.).

    .. note:: If you use Mopidy, and unless you have quite specific use-cases
        (like you don't want to expose the Mopidy HTTP interface, or you have
        some legacy automation that uses the MPD interface), you should use the
        :class:`platypush.plugins.music.mopidy.MusicMopidyPlugin` plugin instead
        of this. The Mopidy plugin provides a more complete and feature-rich
        experience, as not all the features of Mopidy are available through the
        MPD interface, and its API is 100% compatible with this plugin. Also,
        this plugin operates a synchronous/polling logic because of the
        limitations of the MPD protocol, while the Mopidy plugin, as it uses the
        Mopidy Websocket API, can operate in a more efficient way and provide
        real-time updates.

    .. note:: As of Mopidy 3.0 MPD is an optional interface provided by the
        ``mopidy-mpd`` extension. Make sure that you have the extension
        installed and enabled on your instance to use this plugin if you want to
        use it with Mopidy instead of MPD.

    """

    _client_lock = threading.RLock()

    def __init__(
        self,
        host: str,
        port: int = 6600,
        poll_interval: Optional[float] = 20.0,
        **kwargs,
    ):
        """
        :param host: MPD IP/hostname.
        :param port: MPD port (default: 6600).
        :param poll_interval: Polling interval in seconds. If set, the plugin
            will poll the MPD server for status updates and trigger change
            events when required. Default: 20 seconds.
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self.conf = MpdConfig(host=host, port=port)
        self.client = None

    def _connect(self, n_tries: int = 2):
        import mpd

        with self._client_lock:
            if self.client:
                return self.client

            error = None
            while n_tries > 0:
                try:
                    n_tries -= 1
                    self.client = mpd.MPDClient()
                    self.client.connect(self.conf.host, self.conf.port)
                    return self.client
                except Exception as e:
                    error = e
                    self.logger.warning(
                        'Connection exception: %s%s',
                        e,
                        (': Retrying' if n_tries > 0 else ''),
                    )
                    self.wait_stop(0.5)

        self.client = None
        if error:
            raise error

        return self.client

    def _exec(self, method: str, *args, **kwargs):
        error = None
        n_tries = int(kwargs.pop('n_tries')) if 'n_tries' in kwargs else 2
        return_status = (
            kwargs.pop('return_status') if 'return_status' in kwargs else True
        )

        while n_tries > 0:
            try:
                self._connect()
                n_tries -= 1
                with self._client_lock:
                    response = getattr(self.client, method)(*args, **kwargs)

                if return_status:
                    return self._status()

                return response
            except Exception as e:
                error = str(e)
                self.logger.warning(
                    'Exception while executing MPD method %s: %s', method, error
                )
                self.client = None

        return None, error

    @staticmethod
    def _dump_directory(item: dict):
        item['type'] = 'directory'
        item['uri'] = item['name'] = item['directory']
        return item

    @staticmethod
    def _dump_playlist(item: dict):
        item['type'] = 'playlist'
        item['uri'] = item['name'] = item['playlist']
        item['last_modified'] = item.pop('last-modified', None)
        return item

    @staticmethod
    def _dump_track(item: dict, pos: Optional[int] = None):
        item['type'] = 'track'
        item['uri'] = item['file']
        for attr in ('time', 'track', 'disc'):
            item[attr] = int(item[attr]) if item.get(attr) is not None else None

        if 'pos' in item:
            item['playlist_pos'] = int(item.pop('pos'))
        elif pos is not None:
            item['playlist_pos'] = pos

        return item

    @action
    def play(self, resource: Optional[str] = None, **__):
        """
        Play a resource by path/URI.

        :param resource: Resource path/URI
        :type resource: str
        """

        if resource:
            self.add(resource, position=0)
            return self.play_pos(0)

        return self._exec('play')

    @action
    def play_pos(self, pos: int):
        """
        Play a track in the current playlist by position number.

        :param pos: Position number.
        """

        return self._exec('play', pos)

    @action
    def pause(self, *_, **__):
        """Pause playback"""

        status = self._status()['state']
        return self._exec('pause') if status == 'play' else self._exec('play')

    @action
    def pause_if_playing(self):
        """Pause playback only if it's playing"""
        status = self._status()['state']
        return self._exec('pause') if status == 'play' else None

    @action
    def play_if_paused(self):
        """Play only if it's paused (resume)"""
        status = self._status()['state']
        return self._exec('play') if status == 'pause' else None

    @action
    def play_if_paused_or_stopped(self):
        """Play only if it's paused or stopped"""
        status = self._status()['state']
        return self._exec('play') if status in ('pause', 'stop') else None

    @action
    def stop(self, *_, **__):  # type: ignore
        """Stop playback"""
        # Note: stop could be interpreted both as "stop the playback" and "stop
        # the plugin". If the stop event is set, we assume that the user wants
        # to stop the plugin.
        if self.should_stop():
            if self.client:
                self.client.disconnect()

            return None

        return self._exec('stop')

    @action
    def play_or_stop(self):
        """Play or stop (play state toggle)"""
        status = self._status()['state']
        if status == 'play':
            return self._exec('stop')
        return self._exec('play')

    @action
    def next(self, *_, **__):
        """Play the next track"""
        return self._exec('next')

    @action
    def previous(self, **__):
        """Play the previous track"""
        return self._exec('previous')

    @action
    def set_volume(self, volume: int, **__):
        """
        Set the volume.

        :param volume: Volume value (range: 0-100).
        """
        return self._exec('setvol', str(volume))

    @action
    def volup(self, step: Optional[float] = None, **kwargs):
        """
        Turn up the volume.

        :param step: Volume up step (default: 5%).
        """
        step = step or kwargs.get('delta') or 5
        volume = int(self._status()['volume'])
        new_volume = min(volume + step, 100)
        return self.set_volume(new_volume)

    @action
    def voldown(self, step: Optional[float] = None, **kwargs):
        """
        Turn down the volume.

        :param step: Volume down step (default: 5%).
        """
        step = step or kwargs.get('delta') or 5
        volume = int(self._status()['volume'])
        new_volume = max(volume - step, 0)
        return self.set_volume(new_volume)

    def _toggle(self, key: str, value: Optional[bool] = None):
        if value is None:
            value = bool(self._status()[key])
        return self._exec(key, int(value))

    @action
    def random(self, value: Optional[bool] = None):
        """
        Set random mode.

        :param value: If set, set the random state this value (true/false).
            Default: None (toggle current state).
        """
        return self._toggle('random', value)

    @action
    def consume(self, value: Optional[bool] = None):
        """
        Set consume mode.

        :param value: If set, set the consume state this value (true/false).
            Default: None (toggle current state)
        """
        return self._toggle('consume', value)

    @action
    def single(self, value: Optional[bool] = None):
        """
        Set single mode.

        :param value: If set, set the consume state this value (true/false).
            Default: None (toggle current state)
        """
        return self._toggle('single', value)

    @action
    def repeat(self, value: Optional[bool] = None):
        """
        Set repeat mode.

        :param value: If set, set the repeat state this value (true/false).
            Default: None (toggle current state)
        """
        return self._toggle('repeat', value)

    @action
    def shuffle(self):
        """
        Shuffles the current playlist.
        """
        return self._exec('shuffle')

    @action
    def save(self, name: str):
        """
        Save the current tracklist to a new playlist with the specified name.

        :param name: Name of the playlist
        """
        return self._exec('save', name)

    @action
    def add(self, resource: str, *_, position: Optional[int] = None, **__):
        """
        Add a resource (track, album, artist, folder etc.) to the current
        playlist.

        :param resource: Resource path or URI.
        :param position: Position where the track(s) will be inserted (default:
            end of the playlist).
        """

        if isinstance(resource, list):
            for r in resource:
                r = self._parse_resource(r)
                try:
                    if position is None:
                        self._exec('add', r)
                    else:
                        self._exec('addid', r, position)
                except Exception as e:
                    self.logger.warning('Could not add %s: %s', r, e)

            return self.status().output

        r = self._parse_resource(resource)

        if position is None:
            return self._exec('add', r)
        return self._exec('addid', r, position)

    @action
    def delete(self, positions):
        """
        Delete the playlist item(s) in the specified position(s).

        :param positions: Positions of the tracks to be removed
        :type positions: list[int]

        :return: The modified playlist
        """
        for pos in sorted(positions, key=int, reverse=True):
            self._exec('delete', pos)
        return self.get_tracks()

    @action
    def delete_playlist(self, playlist: Union[str, Collection[str]]):
        """
        Permanently remove playlist(s) by name

        :param playlist: Name or list of playlist names to remove
        :type playlist: str or list[str]
        """
        if isinstance(playlist, str):
            playlist = [playlist]
        elif not isinstance(playlist, list):
            raise RuntimeError(f'Invalid type for playlist: {type(playlist)}')

        for p in playlist:
            self._exec('rm', p)

    @action
    def move(
        self,
        start: Optional[int] = None,
        end: Optional[int] = None,
        position: Optional[int] = None,
        from_pos: Optional[int] = None,
        to_pos: Optional[int] = None,
    ):
        """
        Move the playlist items from the positions ``start`` to ``end`` to the
        new position ``position``.

        You can pass either:

            - ``start``, ``end`` and ``position`` to move a slice of tracks
              from ``start`` to ``end`` to the new position ``position``.
            - ``from_pos`` and ``to_pos`` to move a single track from
              ``from_pos`` to ``to_pos``.

        .. note: Positions are 0-based (i.e. the first track has position 0).

        :param start: Start position of the selection.
        :param end: End position of the selection.
        :param position: New position.
        :param from_pos: Alias for ``start`` - it only works with one track at
            the time.
        :param to_pos: Alias for ``position`` - it only works with one track at
            the time.
        """
        assert (start is not None and end is not None and position is not None) or (
            from_pos is not None and to_pos is not None
        ), 'Specify either (start, end, position) or (from_pos, to_pos)'

        if from_pos is not None and to_pos is not None:
            return self._exec('move', from_pos, to_pos)

        chunk = start if start == end else f'{start}:{end}'
        return self._exec('move', chunk, position)

    @classmethod
    def _parse_resource(cls, resource):
        if not resource:
            return None

        m = re.search(r'^https?://open\.spotify\.com/([^?]+)', resource)
        if m:
            resource = 'spotify:' + m.group(1).replace('/', ':')

        if resource.startswith('spotify:'):
            resource = resource.split('?')[0]

        m = re.match(r'spotify:playlist:(.*)', resource)
        if m:
            # Old Spotify URI format, convert it to new
            resource = 'spotify:user:spotify:playlist:' + m.group(1)
        return resource

    @action
    def load(self, playlist, play=True):
        """
        Load and play a playlist by name

        :param playlist: Playlist name
        :type playlist: str

        :param play: Start playback after loading the playlist (default: True)
        :type play: bool
        """

        ret = self._exec('load', playlist)
        if play:
            self.play()
        return ret

    @action
    def clear(self, **__):
        """Clear the current playlist"""
        return self._exec('clear')

    @action
    def seek(self, position: float, **__):
        """
        Seek to the specified position

        :param position: Seek position in seconds, or delta string (e.g. '+15'
            or '-15') to indicate a seek relative to the current position
        """
        return self._exec('seekcur', position)

    @action
    def forward(self):
        """Go forward by 15 seconds"""
        return self._exec('seekcur', '+15')

    @action
    def back(self):
        """Go backward by 15 seconds"""
        return self._exec('seekcur', '-15')

    def _status(self) -> dict:
        n_tries = 2
        error = None

        while n_tries > 0:
            try:
                n_tries -= 1
                self._connect()
                if self.client:
                    return self.client.status()  # type: ignore
            except Exception as e:
                error = e
                self.logger.warning('Exception while getting MPD status: %s', e)
                self.client = None

        raise AssertionError(str(error))

    @action
    def status(self, *_, **__):
        """
        :returns: The current state.

        Example response::

            output = {
                "volume": "9",
                "repeat": "0",
                "random": "0",
                "single": "0",
                "consume": "0",
                "playlist": "52",
                "playlistlength": "14",
                "xfade": "0",
                "state": "play",
                "song": "9",
                "songid": "3061",
                "nextsong": "10",
                "nextsongid": "3062",
                "time": "161:255",
                "elapsed": "161.967",
                "bitrate": "320"
            }

        """
        return self._status()

    def _current_track(self):
        track = self._exec('currentsong', return_status=False)
        if not isinstance(track, dict):
            return None

        if 'title' in track and (
            'artist' not in track
            or not track['artist']
            or re.search('^https?://', track['file'])
            or re.search('^tunein:', track['file'])
        ):
            m = re.match(r'^\s*(.+?)\s+-\s+(.*)\s*$', track['title'])
            if m and m.group(1) and m.group(2):
                track['artist'] = m.group(1)
                track['title'] = m.group(2)

        return track

    @action
    def currentsong(self):
        """
        Legacy alias for :meth:`.current_track`.
        """
        return self.current_track()

    @action
    def current_track(self, *_, **__):
        """
        :returns: The currently played track.

        Example response::

            output = {
                "file": "spotify:track:7CO5ADlDN3DcR2pwlnB14P",
                "time": "255",
                "artist": "Elbow",
                "album": "Little Fictions",
                "title": "Kindling",
                "date": "2017",
                "track": "10",
                "pos": "9",
                "id": "3061",
                "albumartist": "Elbow",
                "x-albumuri": "spotify:album:6q5KhDhf9BZkoob7uAnq19"
            }
        """
        return self._current_track()

    @action
    def get_tracks(self, *_, **__):
        """
        :returns: The tracks in the current playlist as a list of dicts.

        Example output::

            output = [
                {
                    "file": "spotify:track:79VtgIoznishPUDWO7Kafu",
                    "time": "355",
                    "artist": "Elbow",
                    "album": "Little Fictions",
                    "title": "Trust the Sun",
                    "date": "2017",
                    "track": "3",
                    "pos": "10",
                    "id": "3062",
                    "albumartist": "Elbow",
                    "x-albumuri": "spotify:album:6q5KhDhf9BZkoob7uAnq19"
                },
                {
                    "file": "spotify:track:3EzTre0pxmoMYRuhJKMHj6",
                    "time": "219",
                    "artist": "Elbow",
                    "album": "Little Fictions",
                    "title": "Gentle Storm",
                    "date": "2017",
                    "track": "2",
                    "pos": "11",
                    "id": "3063",
                    "albumartist": "Elbow",
                    "x-albumuri": "spotify:album:6q5KhDhf9BZkoob7uAnq19"
                },
            ]
        """
        return [
            self._dump_track(track, pos=i)  # type: ignore
            for i, track in enumerate(self._exec('playlistinfo', return_status=False))
        ]

    @action
    def get_playlists(self, *_, **__):
        """
        :returns: The playlists available on the server as a list of dicts.

        Example response::

            output = [
                {
                    "playlist": "Rock",
                    "last_modified": "2018-06-25T21:28:19Z"
                },
                {
                    "playlist": "Jazz",
                    "last_modified": "2018-06-24T22:28:29Z"
                },
                {
                    # ...
                }
            ]

        """
        playlists: list = self._exec(  # type: ignore
            'listplaylists', return_status=False
        )

        return sorted(
            [self._dump_playlist(pl) for pl in playlists], key=lambda p: p['playlist']
        )

    @action
    def get_playlist(self, playlist: str, *_, **__):
        """
        Get the tracks in a playlist.

        :param playlist: Name of the playlist
        """
        return [
            self._dump_track(track)  # type: ignore
            for track in self._exec(
                'listplaylistinfo',  # if with_tracks else 'listplaylist',
                playlist,
                return_status=False,
            )
            if track
        ]

    @action
    def add_to_playlist(
        self, playlist: str, resources: Union[str, Collection[str]], **_
    ):
        """
        Add one or multiple resources to a playlist.

        :param playlist: Playlist name
        :param resources: URI or path of the resource(s) to be added
        """

        if isinstance(resources, str):
            resources = [resources]

        for res in resources:
            self._exec('playlistadd', playlist, res)

    @action
    def remove_from_playlist(
        self, playlist: str, resources: Union[int, Collection[int]], *_, **__
    ):
        """
        Remove one or multiple tracks from a playlist.

        :param playlist: Playlist name
        :param resources: Position or list of positions to remove
        """

        if isinstance(resources, str):
            resources = int(resources)
        if isinstance(resources, int):
            resources = [resources]

        for p in sorted(resources, reverse=True):
            self._exec('playlistdelete', playlist, p)

    @action
    def playlist_move(self, playlist: str, from_pos: int, to_pos: int, *_, **__):
        """
        Change the position of a track in the specified playlist.

        :param playlist: Playlist name
        :param from_pos: Original track position
        :param to_pos: New track position
        """
        self._exec('playlistmove', playlist, from_pos, to_pos)

    @action
    def playlist_clear(self, name: str):
        """
        Clears all the elements from the specified playlist.

        :param name: Playlist name.
        """
        self._exec('playlistclear', name)

    @action
    def rename_playlist(self, playlist: str, new_name: str):
        """
        Rename a playlist.

        :param playlist: Original playlist name or URI
        :param new_name: New playlist name
        """
        self._exec('rename', playlist, new_name)

    @action
    def browse(self, uri: Optional[str] = None):
        """
        Browse the items under the specified URI.

        :param uri: URI to browse (default: root directory).
        """
        resp: dict = (  # type: ignore
            self._exec('lsinfo', uri, return_status=False)
            if uri
            else self._exec('lsinfo', return_status=False)
        )

        ret = []
        for item in resp:
            if item.get('directory'):
                item = self._dump_directory(item)
            elif item.get('playlist'):
                item = self._dump_playlist(item)
            elif item.get('file'):
                item = self._dump_track(item)
            else:
                continue

            ret.append(item)

        return ret

    @action
    def plchanges(self, version: int):
        """
        Show what has changed on the current playlist since a specified playlist
        version number.

        :param version: Version number
        :returns: A list of dicts representing the songs being added since the specified version
        """
        return self._exec('plchanges', version, return_status=False)

    @action
    def searchaddplaylist(self, name: str):
        """
        Search and add a playlist by (partial or full) name.

        :param name: Playlist name, can be partial.
        """

        resp: list = self._exec('listplaylists', return_status=False)  # type: ignore
        playlists = [
            pl['playlist'] for pl in resp if name.lower() in pl['playlist'].lower()
        ]

        if not playlists:
            return None

        self._exec('clear')
        self._exec('load', playlists[0])
        self._exec('play')
        return {'playlist': playlists[0]}

    @staticmethod
    def _make_filter(f: dict) -> list:
        ll = []
        for k, v in f.items():
            ll.extend([k, v])
        return ll

    @action
    def find(self, filter: dict, *args, **kwargs):  # pylint: disable=redefined-builtin
        """
        Find in the database/library by filter.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """
        filter_list = self._make_filter(filter)
        return self._exec('find', *filter_list, *args, return_status=False, **kwargs)

    @action
    def findadd(
        self, filter: dict, *args, **kwargs  # pylint: disable=redefined-builtin
    ):
        """
        Find in the database/library by filter and add to the current playlist.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """
        filter_list = self._make_filter(filter)
        return self._exec('findadd', *filter_list, *args, return_status=False, **kwargs)

    @action
    def search(
        self,
        *args,
        query: Optional[Union[str, dict]] = None,
        filter: Optional[dict] = None,  # pylint: disable=redefined-builtin
        **kwargs,
    ):
        """
        Free search by filter.

        :param query: Free-text query or search structured filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``).
        :param filter: Structured search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``) - same as
            ``query``, it's still here for back-compatibility reasons.
        :returns: list[dict]
        """
        assert query or filter, 'Specify either `query` or `filter`'

        filt = filter
        if isinstance(query, str):
            filt = query
        elif isinstance(query, dict):
            filt = {**(filter or {}), **query}

        filter_list = self._make_filter(filt) if isinstance(filt, dict) else [query]

        items: list = self._exec(  # type: ignore
            'search', *filter_list, *args, return_status=False, **kwargs
        )

        # Spotify results first
        return sorted(
            items, key=lambda item: 0 if item['file'].startswith('spotify:') else 1
        )

    @action
    def searchadd(
        self, filter: dict, *args, **kwargs  # pylint: disable=redefined-builtin
    ):
        """
        Free search by filter and add the results to the current playlist.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """
        filter_list = self._make_filter(filter)
        return self._exec(
            'searchadd', *filter_list, *args, return_status=False, **kwargs
        )

    def main(self):
        listener = None
        if self.poll_interval is not None:
            listener = MpdListener(self)
            listener.start()

        self.wait_stop()

        if listener:
            listener.join()


# vim:sw=4:ts=4:et:



import logging
import os
import signal
import subprocess
import sys
import time

from platypush.process import ControllableProcess


class ApplicationProcess(ControllableProcess):
    """
    Controllable process wrapper interface for the main application.
    """

    def __init__(self, *args: str, pidfile: str, **kwargs):
        super().__init__(name='platypush', **kwargs)

        self.logger = logging.getLogger('platypush')
        self.args = args
        self.pidfile = pidfile

    def __enter__(self) -> "ApplicationProcess":
        self.start()
        return self

    def __exit__(self, *_, **__):
        self.stop()

    def main(self):
        self.logger.info('Starting application...')
        app = None

        try:
            with subprocess.Popen(
                [sys.executable, '-m', 'platypush.app', *self.args],
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
            ) as app:
                app.wait()
        except KeyboardInterrupt:
            pass

        if app and app.poll() is None:
            app.terminate()

        wait_start = time.time()
        while app and app.poll() is None:
            if time.time() - wait_start > 5:
                self.logger.warning('Application did not terminate, killing it')
                app.kill()
                break

            time.sleep(0.1)

    def on_stop(self):
        try:
            with open(self.pidfile, 'r') as f:
                pid = int(f.read().strip())
        except (OSError, ValueError):
            pid = None

        if not pid:
            return

        try:
            os.kill(pid, signal.SIGINT)
        except OSError:
            pass
